/*-
 * Copyright (c) 2007 Akihiko GOTANDA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * Driver for SD memory card & MMC
 */

#include <dev/sdmmc/sdvar.h>
#include <dev/sdmmc/sdshcreg.h>

#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/bio.h>
#include <geom/geom_disk.h>

#include "sdbus_if.h"


struct sd_softc {
    device_t	dev;
    device_t	pdev;

    struct mtx	mtx;
    int		locked;

    int		slot_no;
    int		card_type;
    int		block_size;
    u_int32_t	card_capacity;	/* unit: 512B */
    int		rca;

    struct disk	*disk;
    int		disk_open_count;

    int		flag;
#define	SD_DISK_CREATED		0x01

    int		error;
};

static int	sd_probe(device_t);
static int	sd_attach(device_t);
static int	sd_detach(device_t);

static device_method_t sd_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		sd_probe),
	DEVMETHOD(device_attach,	sd_attach),
	DEVMETHOD(device_detach,	sd_detach),

	{ 0, 0 }
};

static driver_t sd_driver = {
	"sd",
	sd_methods,
	sizeof(struct sd_softc)
};

static devclass_t sd_devclass;

DRIVER_MODULE(sd, sdshcbus, sd_driver, sd_devclass, NULL, NULL);
MODULE_VERSION(sd, 1);


static disk_open_t	sd_open;
static disk_strategy_t	sd_strategy;
static disk_close_t	sd_close;


struct mid {
    int		mid;
    const char	*manufacturer;
};

static struct mid sd_mid_list[], mmc_mid_list[];


static int
sd_probe(device_t dev)
{
    struct sdshcbus_softc *psc = device_get_softc(device_get_parent(dev));
    struct mid *mid, *mid_list;
    const char *manufacturer = "Unknown";
    char desc[128];

    if (is_card_mmc(psc->card_inserted)) {
	mid_list = mmc_mid_list;
    } else {
	mid_list = sd_mid_list;
    }
    for (mid = mid_list; mid->manufacturer != NULL; mid++) {
	if (psc->mid == mid->mid) {
	    manufacturer = mid->manufacturer;
	    break;
	}
    }
    if (is_card_sdmem(psc->card_inserted) || is_card_mmc(psc->card_inserted)) {
	sprintf(desc, "%s %s %s rev %d.%d",
		manufacturer, psc->oid, psc->pnm, psc->prv[0], psc->prv[1]);
	device_set_desc_copy(dev, desc);
	return 0;
    }

    return ENXIO;
}


static int
sd_attach(device_t dev)
{
    device_t pdev = device_get_parent(dev);
    struct sdshcbus_softc *psc = device_get_softc(pdev);
    struct sd_softc *sc = device_get_softc(dev);
    int max_clock = psc->card_max_clock;
    int card_clock = psc->card_clock;
    int card_kbps = card_clock * psc->card_buswidth;
    u_int32_t cap;
    u_int64_t access_time;
    int rdtm_factor, rdtm_unit, wrtm_factor, wrtm_unit;

    sc->dev = dev;
    sc->pdev = pdev;

    mtx_init(&sc->mtx, device_get_nameunit(dev), NULL,
	     MTX_DEF | MTX_RECURSE);
    sc->locked = 0;

    sc->slot_no = device_get_unit(pdev);
    sc->card_type = psc->card_inserted;

    sc->block_size = psc->block_size;
    sc->card_capacity = psc->card_capacity;
    sc->rca = psc->rca;

    switch (sc->card_type) {
      case SDDEV_SD1:
	MSG((sc->dev, "Ver 1.x SD memory card\n"));
	break;
      case SDDEV_SD2:
	MSG((sc->dev, "Ver 2.00 Standard Capacity SD memory card\n"));
	break;
      case SDDEV_SDHC:
	MSG((sc->dev, "Ver 2.00 High Capacity SD memory card\n"));
	break;
      case SDDEV_MMC:
	MSG((sc->dev, "Multi Media Card\n"));
	break;
      case SDDEV_MMCHD:
	MSG((sc->dev, "High Density Multi Media Card\n"));
	break;
    }
    cap = sc->card_capacity >> 1; /* unit: kB */
    MSG((sc->dev, "%u%cB (%u x %d-byte blocks)\n",
	 cap < 10240 ? cap : cap / 1024, cap < 10240 ? 'k' : 'M',
	 (u_int32_t)(((u_int64_t)sc->card_capacity << 9) / sc->block_size),
	 sc->block_size));
    access_time = psc->card_read_access_time;
    if (10000000 <= access_time) {
	rdtm_factor = (int)(access_time / 1000000);
	rdtm_unit = 'm';
    } else if (10000 <= access_time) {
	rdtm_factor = (int)(access_time / 1000);
	rdtm_unit = 'u';
    } else {
	rdtm_factor = (int)access_time;
	rdtm_unit = 'n';
    }
    access_time *= psc->card_write_speed_factor;
    if (10000000 <= access_time) {
	wrtm_factor = (int)(access_time / 1000000);
	wrtm_unit = 'm';
    } else if (10000 <= access_time) {
	wrtm_factor = (int)(access_time / 1000);
	wrtm_unit = 'u';
    } else {
	wrtm_factor = (int)access_time;
	wrtm_unit = 'n';
    }
    MSG((sc->dev, "access time: read %d.%d%cs, write %d.%d%cs\n",
	 rdtm_factor / 10, rdtm_factor % 10, rdtm_unit,
	 wrtm_factor / 10, wrtm_factor % 10, wrtm_unit));
    if (0 <= psc->card_speed_class) {
	MSG((sc->dev, "speed class %d\n", psc->card_speed_class));
    }
    MSG((sc->dev, "max clock %d%cHz, drive at %d.%03d%cHz\n",
	 (1000 <= max_clock ? max_clock / 1000 : max_clock),
	 (1000 <= max_clock ? 'M' : 'k'),
	 (1000 <= card_clock ? card_clock / 1000 : card_clock),
	 (1000 <= card_clock ? card_clock % 1000 : 0),
	 (1000 <= card_clock ? 'M' : 'k')));
    MSG((sc->dev, "%d.%03d%cB/s transfers with %d-bit bus width\n",
	 (8000 <= card_kbps ? card_kbps / 8000 : card_kbps),
	 (8000 <= card_kbps ? (card_kbps / 8) % 1000 : 0),
	 (8000 <= card_kbps ? 'M' : 'k'),
	 psc->card_buswidth));

    sc->flag = 0;
    sc->error = 0;

    sc->disk_open_count = 0;
    LOG(0x08, (sc->dev, "sd_attach: creating GEOM\n"));
    /* create the disk device */
    sc->disk = disk_alloc();
/*    sc->disk->d_flags = DISKFLAG_NEEDSGIANT | DISKFLAG_CANDELETE; */
    sc->disk->d_flags = DISKFLAG_CANDELETE;
    if (is_card_sdmem(sc->card_type)) {
	sc->disk->d_name = "sd";
    } else {
	sc->disk->d_name = "mmc";
    }
    sc->disk->d_unit = device_get_unit(sc->dev);
    sc->disk->d_open = sd_open;
    sc->disk->d_close = sd_close;
    sc->disk->d_strategy = sd_strategy;
/*    sc->disk->d_ioctl = sd_ioctl; */
/*    sc->disk->d_dump = sd_dump; */
    sc->disk->d_sectorsize = sc->block_size;
    sc->disk->d_mediasize = (u_int64_t)sc->card_capacity << 9;
    sc->disk->d_maxsize = sc->block_size * psc->max_block_count;

    if (sc->disk->d_mediasize < sc->disk->d_maxsize) {
	sc->disk->d_maxsize = sc->disk->d_mediasize;
    }
    sc->disk->d_drv1 = sc;
    disk_create(sc->disk, DISK_VERSION);
    sc->flag |= SD_DISK_CREATED;

    return 0;
}


static int
sd_detach(device_t dev)
{
    struct sd_softc *sc = device_get_softc(dev);

    LOG(0x10, (sc->dev, "sd_detach\n"));
    /* check that we have a valid disk to detach */
    if (sc->disk != NULL) {
	/* detroy disk from the system */
	disk_gone(sc->disk);
	disk_destroy(sc->disk);
	sc->disk = NULL;
    }
    sc->flag &= ~SD_DISK_CREATED;

    return 0;
}


static int
sd_open(struct disk *dp)
{
    struct sd_softc *sc = (struct sd_softc *)dp->d_drv1;

    if (sc == NULL) {
	printf("sd_open: Can't get softc\n");
	return (ENXIO);
    }

    LOG(0x08, (sc->dev, "sd_open #%d\n", sc->slot_no));

    if (sc->disk_open_count == 0) {
#if 0
	/* Addressing card */
	if (SDBUS_IO_REQUEST(sc->pdev, 7, sc->rca, 0, 0, 0, NULL)) {
	    MSG((sc->dev, "CMD7 (SELECT_CARD: 0x%08x) failed\n",
		 sc->rca));
	    return (EIO);
	}

	if (sc->card_type != SDDEV_SDHC) {
	    /* Set block size */
	    int blk_size = sc->disk->d_sectorsize;
	    if (SDBUS_IO_REQUEST(sc->pdev, 16, blk_size, 0, 0, 0, NULL)) {
		MSG((sc->dev, "CMD16 (SET_BLOCKLEN: %d) failed\n",
		     blk_size));
		return ENXIO;
	    }
	}
#endif
    }

    ++sc->disk_open_count;
    sc->flag |= SDSHC_STAT_OPENED;
    return 0;
}


static int
sd_close(struct disk *dp)
{
    struct sd_softc *sc = (struct sd_softc *)dp->d_drv1;

    if (sc == NULL) {
	printf("sd_close: Can't get softc\n");
	return (ENXIO);
    }

    LOG(0x08, (sc->dev, "sd_close #%d\n", sc->slot_no));

    if (0 < sc->disk_open_count) {
	if (--sc->disk_open_count == 0) {
#if 0
	    /* De-select card */
	    if (SDBUS_IO_REQUEST(sc->pdev, 7, 0, 0, 0, 0, NULL)) {
		MSG((sc->dev, "CMD7 (DESELECT_CARD) failed\n"));
	    }
#endif

	    sc->flag &= ~SDSHC_STAT_OPENED;
	}
    }
    return 0;
}


static void
sd_strategy(struct bio *bp)
{
    struct sd_softc *sc = (struct sd_softc *)bp->bio_disk->d_drv1;
    int blk_size, blk_count, write = 0, cmd;
    u_int32_t offset;

    if (sc == NULL) {
	printf("sd_strategy: Can't get softc\n");
	biofinish(bp, NULL, ENXIO);
	return;
    }

    blk_size = sc->disk->d_sectorsize;
    blk_count = bp->bio_bcount / blk_size;
    offset = bp->bio_offset;

    LOG(0x08, (sc->dev, "sd_strategy #%d: %d:%d:%d\n",
	       sc->slot_no, offset, blk_size, blk_count));

    if (blk_count == 0) {
	/* block count is 0; do nothing */
	bp->bio_resid = bp->bio_bcount;
	biodone(bp);
	return;
    }
/* - DO NOT USE LOCK HERE
    SDC_LOCK(sc);
*/
#if 0
    /* Addressing card */
    if (SDBUS_IO_REQUEST(sc->pdev, 7, sc->rca, 0, 0, 0, NULL)) {
	MSG((sc->dev, "CMD7 (SELECT_CARD: 0x%08x) failed\n",
	     sc->rca));
	biofinish(bp, NULL, EIO);
    }
#endif

    if (sc->card_type == SDDEV_SDHC) {
	/* in SDHC, data address is in block (512B) units */
	offset >>= 9;
    }

    switch (bp->bio_cmd) {
      case BIO_READ:
	if (blk_count == 1) {
	    cmd = 17;	/* READ_SINGLE_BLOCK */
	} else {
	    cmd = 18;	/* READ_MULTIPLE_BLOCK */
	}
	break;
      case BIO_WRITE:
	if (blk_count == 1) {
	    cmd = 24;	/* WRITE_BLOCK */
	} else {
	    cmd = 25;	/* WRITE_MULTIPLE_BLOCK */
	}
	write = 1;
	break;
      default:
	MSG((sc->dev, "FAILURE - unknown BIO operation\n"));
	biofinish(bp, NULL, EIO);
/* - DO NOT USE LOCK HERE
	SD_UNLOCK(sc);
*/
	return;
    }

    LOG(0x10, (sc->dev,
	       "I/O %c, %d:%d:%d\n",
	       write ? 'W' : 'R', offset, blk_size, blk_count));
    if (!sc->error) {
#if defined(SD_TRANS_USE_POLLING)
	sdc_disable_intr(sc);
	sc->use_polling = 1;
	sc->result = 0;
	sdc_enable_intr_polling(sc);
	/*msleep(sc, &sc->mtx, PPAUSE, "sd_strategy", 1);*/
#endif
	if ((sc->error = SDBUS_IO_REQUEST(sc->pdev, cmd, offset, write,
					 blk_size, blk_count, bp->bio_data))) {
	    MSG((sc->dev, "CMD%d <%c, %d:%d:%d> failed\n",
		 cmd, write ? 'W' : 'R', offset, blk_size, blk_count));
	    biofinish(bp, NULL, EIO);
	}
#if defined(SD_TRANS_USE_POLLING)
	sc->use_polling = 0;
	sdc_enable_intr(sc);
#endif
	if (sc->error) {
	    bp->bio_error = EIO;
	    bp->bio_flags |= BIO_ERROR;
	}
    } else {
	/* dummy R/W - do nothing */
    }

    bp->bio_resid = bp->bio_bcount - blk_size * blk_count;

    biodone(bp);
/* - DO NOT USE LOCK HERE
    SD_UNLOCK(sc);
*/
}


static struct mid sd_mid_list[] = {
    { 0x01,	"Panasonic"	},
    { 0x02,	"Toshiba"	},
    { 0x03,	"SanDisk"	},

    { 0x13,	"KINGMAX"	},
/*    { 0x15,	"Lexar"		},*/
    { 0x18,	"Infineon"	},
    { 0x1a,	"PQI"		},
    { 0x1b,	"Transcend"	},
    { 0x1d,	"A-DATA"	},
    { 0x27,	"Patriot"	},
    { 0x6f,	"TECIC"		},

    { 0,	NULL		}
};


static struct mid mmc_mid_list[] = {
    { 0x02,	"SanDisk"	},
    { 0x06,	"Renesas"	},
    { 0x07,	"Macronix"	},
    { 0x15,	"Samsung"	},
    { 0x16,	"Matrix"	},
    { 0x37,	"KINGMAX"	},
    { 0x41,	"OKI"		},

    { 0,	NULL		}
};
