/*-
 * Copyright (c) 2007 Akihiko GOTANDA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * Driver for SD card slot
 */

#include <dev/sdmmc/sdvar.h>
#include <dev/sdmmc/sdshcreg.h>

#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/malloc.h>
#include <sys/ctype.h>
#include <machine/bus.h>

#include "sdbus_if.h"


static int	sdshcbus_probe(device_t);
static int	sdshcbus_attach(device_t);
static int	sdshcbus_detach(device_t);
static int	sdshcbus_suspend(device_t);
static int	sdshcbus_resume(device_t);

static int	sdshcbus_intr(device_t);

static int	sdshcbus_io_request(device_t dev,
				    u_int32_t cmd_index, u_int32_t arg,
				    int write, int blk_size, int num_blk,
				    void *data);
static int	sdshcbus_acmd_io_request(device_t dev,
					 u_int32_t acmd_index, u_int32_t arg,
					 int write, int blk_size, int num_blk,
					 void *data);


static device_method_t sdshcbus_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,			sdshcbus_probe),
	DEVMETHOD(device_attach,		sdshcbus_attach),
	DEVMETHOD(device_detach,		sdshcbus_detach),
	DEVMETHOD(device_suspend,		sdshcbus_suspend),
	DEVMETHOD(device_resume,		sdshcbus_resume),

	/* SDSHC local method */
	DEVMETHOD(sdbus_intr,			sdshcbus_intr),

	DEVMETHOD(sdbus_io_request,		sdshcbus_io_request),
	DEVMETHOD(sdbus_acmd_io_request,	sdshcbus_acmd_io_request),

	{ 0, 0 }
};

static driver_t sdshcbus_driver = {
	"sdshcbus",
	sdshcbus_methods,
	sizeof(struct sdshcbus_softc)
};

static devclass_t sdshcbus_devclass;
#define SDSHCBUS_SOFTC(unit) \
	((struct sdshcbus_softc*)devclass_get_softc(sdshcbus_devclass, unit))

DRIVER_MODULE(sdshcbus, sdshc, sdshcbus_driver, sdshcbus_devclass, NULL, NULL);
MODULE_VERSION(sdshcbus, 1);


struct cmd_type {
    int resp_type;
    int need_transfer;
};

#define	CMD_RESERVED	-1
#define	CMD_NO_RESP	0x00
#define	CMD_R1		0x1a
#define	CMD_R1b		0x1b
#define	CMD_R2		0x09
#define	CMD_R3		0x02
#define	CMD_R4		0x02
#define	CMD_R5		0x1a
#define	CMD_R5b		0x1b
#define	CMD_R6		0x1a
#define	CMD_R7		0x1a

struct cmd_type sd_cmd_attr[64] = {
    /*  0 */	{ CMD_NO_RESP,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R2,	0 },	{ CMD_R6,	0 },
		{ CMD_NO_RESP,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1,	1 },	{ CMD_R1b,	0 },
		{ CMD_R7,	0 },	{ CMD_R2,	0 },
    /* 10 */	{ CMD_R2,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1b,	0 },	{ CMD_R1,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_NO_RESP,	0 },
		{ CMD_R1,	0 },	{ CMD_R1,	1 },
		{ CMD_R1,	1 },	{ CMD_RESERVED,	0 },
    /* 20 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1,	1 },	{ CMD_R1,	1 },
		{ CMD_RESERVED,	0 },	{ CMD_R1,	1 },
		{ CMD_R1b,	0 },	{ CMD_R1b,	0 },
    /* 30 */	{ CMD_R1,	1 },	{ CMD_RESERVED,	0 },
		{ CMD_R1,	0 },	{ CMD_R1,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1b,	0 },	{ CMD_RESERVED,	0 },
    /* 40 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1,	1 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 50 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_R1,	0 },
		{ CMD_R1,	1 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 60 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 }
};

struct cmd_type sd_acmd_attr[64] = {
    /*  0 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_R2,	0 },
    /* 10 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_R1,	1 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	1 },	{ CMD_RESERVED,	0 },
    /* 20 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1,	1 },	{ CMD_R1,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 30 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 40 */	{ CMD_RESERVED,	0 },	{ CMD_R3,	0 },
		{ CMD_R1,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 50 */	{ CMD_RESERVED,	0 },	{ CMD_R1,	1 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 60 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 }
};

struct cmd_type mmc_cmd_attr[64] = {
    /*  0 */	{ CMD_NO_RESP,	0 },	{ CMD_R3,	0 },
		{ CMD_R2,	0 },	{ CMD_R1,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1b,	0 },	{ CMD_R1b,	0 },
		{ CMD_R1,	1 },	{ CMD_R2,	0 },
    /* 10 */	{ CMD_R2,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1b,	0 },	{ CMD_R1,	0 },
		{ CMD_R1,	1 },	{ CMD_NO_RESP,	0 },
		{ CMD_R1,	0 },	{ CMD_R1,	1 },
		{ CMD_R1,	1 },	{ CMD_R1,	1 },
    /* 20 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_R1,	0 },
		{ CMD_R1,	1 },	{ CMD_R1,	1 },
		{ CMD_RESERVED,	0 },	{ CMD_R1,	1 },
		{ CMD_R1b,	0 },	{ CMD_R1b,	0 },
    /* 30 */	{ CMD_R1,	1 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_R1,	0 },
		{ CMD_R1,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1b,	0 },	{ CMD_RESERVED,	0 },
    /* 40 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_R1b,	1 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 50 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
    /* 60 */	{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 },
		{ CMD_RESERVED,	0 },	{ CMD_RESERVED,	0 }
};


#if defined(SD_DEBUG)
static char *card_state[16] = {
    "idle", "ready", "ident", "stby", "tran", "data", "rcv", "prg", "dis",
    "rsvd", "rsvd", "rsvd", "rsvd", "rsvd", "rsvd", "rsvd for I/O mode"
};
#endif


static int	sdshcbus_init(struct sdshcbus_softc *sc);
static int	sdshcbus_config(struct sdshcbus_softc *sc);

static void	sdshcbus_clear_intr(struct sdshcbus_softc *sc);
static void	sdshcbus_disable_intr(struct sdshcbus_softc *sc);
static void	sdshcbus_enable_intr(struct sdshcbus_softc *sc);
static void	sdshcbus_enable_intr_polling(struct sdshcbus_softc *sc);
static void	sdshcbus_intr_handler(struct sdshcbus_softc *sc);
static void	sdshcbus_error_recovery(struct sdshcbus_softc *sc);
static int	sdshcbus_check_intr_clean(struct sdshcbus_softc *sc);
static int	sdshcbus_reset_lines(struct sdshcbus_softc *sc);
static int	sdshcbus_reset_cmd_line(struct sdshcbus_softc *sc);
static int	sdshcbus_reset_dat_line(struct sdshcbus_softc *sc);
static int	sdshcbus_reset(struct sdshcbus_softc *sc);
static void 	sdshcbus_stop(struct sdshcbus_softc *sc);

static void	sdshcbus_card_insertion_callback(void *arg, int count);
static void	sdshcbus_card_removal_callback(void *arg, int count);
static void	sdshcbus_event_card_insertion(struct sdshcbus_softc *sc);
static void	sdshcbus_event_card_removal(struct sdshcbus_softc *sc);

static int	sdshcbus_check_card_inserted(struct sdshcbus_softc *sc);
static int	sdshcbus_check_command_running(struct sdshcbus_softc *sc);
static int	sdshcbus_check_clock_supplied(struct sdshcbus_softc *sc);
static void	sdshcbus_print_card_type(struct sdshcbus_softc *sc);
static int	sdshcbus_led_control(struct sdshcbus_softc *sc, int state);
static int	sdshcbus_clock_supply(struct sdshcbus_softc *sc, int clock);
static int	sdshcbus_clock_stop(struct sdshcbus_softc *sc);
static int	sdshcbus_clock_frequency_change(struct sdshcbus_softc *sc,
						int clock);
static void	sdshcbus_card_powerup(struct sdshcbus_softc *sc);
static void	sdshcbus_card_powerdown(struct sdshcbus_softc *sc);
static int	sdshcbus_card_bootstrap(struct sdshcbus_softc *sc);
static int	sdshcbus_card_init(struct sdshcbus_softc *sc);
static int	sdshcbus_sd_get_CSD(struct sdshcbus_softc *sc);
static int	sdshcbus_card_init_mmc(struct sdshcbus_softc *sc);
static void	sdshcbus_set_bus_width(struct sdshcbus_softc *sc, int width);
static void	sdshcbus_change_bus_width(struct sdshcbus_softc *sc, int width);
static void	sdshcbus_set_highspeed_mode(struct sdshcbus_softc *sc, int enable);
static void	sdshcbus_set_data_timeout(struct sdshcbus_softc *sc,
					  u_int32_t timeout);
static u_int32_t sdshcbus_build_cmd(struct cmd_type *cmd_attr, int index);
static int	sdshcbus_transaction_wo_data(struct sdshcbus_softc *sc,
					     u_int32_t cmd, u_int32_t arg,
					     int with_trans_complete);
static int	sdshcbus_command_complete_sequence(struct sdshcbus_softc *sc,
						   int with_trans_complete);
static int	sdshcbus_transaction_with_data(struct sdshcbus_softc *sc,
					       u_int32_t cmd, u_int32_t arg,
					       int write,
					       int blk_size, int num_blk,
					       void *data);
static int	sdshcbus_transaction(struct sdshcbus_softc *sc,
				     struct cmd_type *cmd_attr,
				     u_int32_t cmd_index, u_int32_t arg,
				     int write, int blk_size, int num_blk,
				     void *data);
#define sdshcbus_sd_transaction(sc, idx, arg, write, blk_size, num_blk, data) \
		sdshcbus_transaction(sc, sd_cmd_attr, idx, arg, \
					write, blk_size, num_blk, data)
#define sdshcbus_mmc_transaction(sc, idx, arg, write, blk_size, num_blk, data) \
		sdshcbus_transaction(sc, mmc_cmd_attr, idx, arg, \
					write, blk_size, num_blk, data)
static int	sdshcbus_acmd_transaction(struct sdshcbus_softc *sc,
					  u_int32_t cmd_index, u_int32_t arg,
					  int write, int blk_size, int num_blk,
					  void *data);
static int	sdshcbus_async_abort(struct sdshcbus_softc *sc);
static int	sdshcbus_sync_abort(struct sdshcbus_softc *sc);


static d_open_t  sdshcbus_open;
static d_close_t sdshcbus_close;
static d_read_t sdshcbus_read;
static d_write_t sdshcbus_write;
/*
static d_poll_t  sdshcbus_poll;
static d_purge_t sdshcbus_purge;
*/

static struct cdevsw sdshcbus_cdevsw = {
    .d_version = D_VERSION,
    .d_flags =	 D_NEEDGIANT,
    .d_open =	 sdshcbus_open,
    .d_close =	 sdshcbus_close,
    .d_read =	 sdshcbus_read,
    .d_write =	 sdshcbus_write,
/*
    .d_poll =	 sdshcbus_poll,
    .d_purge =	 sdshcbus_purge,
*/
    .d_name =	 "sdshcbus",
};

#define	SDSHCBUS_UNIT(dev)	minor(dev)


static int
sdshcbus_probe(device_t dev)
{
    device_set_desc(dev, "SD Standard Host Controller slot bus");
    return 0;
}


static int
sdshcbus_attach(device_t dev)
{
    struct sdshcbus_softc *sc = device_get_softc(dev);
    device_t pdev = device_get_parent(dev);
    int slot = device_get_unit(dev);
    int error;

    sc->dev = dev;

    mtx_init(&sc->mtx, device_get_nameunit(dev), NULL,
	     MTX_DEF | MTX_RECURSE);
    sc->locked = 0;

    LOG(0x10, (sc->dev, "sdshcbus_attach: slot #%d\n", slot));
    sc->slot_no[0] = slot;

    sc->max_block_count = SDSHC_MAX_BLOCK_COUNT;

    if ((error = sdshcbus_init(sc)) == 0) {
	sc->cdev = make_dev(&sdshcbus_cdevsw, slot,
			    UID_ROOT, GID_OPERATOR, 0664,
			    "%s.%d", device_get_nameunit(pdev), slot);

	TASK_INIT(&sc->attach_task, 0, sdshcbus_card_insertion_callback, sc);
	TASK_INIT(&sc->detach_task, 0, sdshcbus_card_removal_callback, sc);

	sc->addr = NULL;
	sc->rest_blk_count = 0;
    }

    return error;
}


static void
sdshcbus_detach_children(device_t dev)
{
    device_t *children;
    int nchildren;
    int cnt;

    /* detach & delete all children */
    bus_generic_detach(dev);
    if (!device_get_children(dev, &children, &nchildren)) {
	for (cnt = 0; cnt < nchildren; cnt++) {
	    device_delete_child(dev, children[cnt]);
	}
	free(children, M_TEMP);
    }
}


static int
sdshcbus_detach(device_t dev)
{
    struct sdshcbus_softc *sc = device_get_softc(dev);
    device_t pdev = device_get_parent(dev);
    struct sdshc_softc *psc = device_get_softc(pdev);
    int *mem_rid = &psc->mem_rid[sc->slot_no[0]];

    LOG(0x10, (sc->dev, "sdshcbus_detach #%d\n", sc->slot_no[0]));
/*
    tsleep(sc, PPAUSE, "sdshcbus_detach", hz);
*/
    sdshcbus_detach_children(dev);

    sdshcbus_stop(sc);

    if (sc->cdev != NULL) {
	destroy_dev(sc->cdev);
	sc->cdev = NULL;
    }

    if (sc->slot_no[0] != 0) {
	bus_release_resource(dev, SYS_RES_MEMORY, *mem_rid, sc->mem);
    }

    return 0;
}


static int
sdshcbus_suspend(device_t dev)
{
    return 0;
}


static int
sdshcbus_resume(device_t dev)
{
    return 0;
}


static int
sdshcbus_intr(device_t dev)
{
    struct sdshcbus_softc *sc = device_get_softc(dev);
/*
    SD_LOCK(sc); - DO NOT LOCK HERE; Lock at sdshc_intr()
*/
    LOG(0x10, (sc->dev, "Intrrupt occured from slot #%d\n", sc->slot_no[1]));

    sdshcbus_disable_intr(sc);

    sdshcbus_intr_handler(sc);

    if (sdshcbus_check_intr_clean(sc)) {
	LOG(0x10,
	    (sc->dev, "Intrrupt for slot #%d is clean\n", sc->slot_no[1]));

	sdshcbus_enable_intr(sc);

    } else {
	LOG(0x10,
	    (sc->dev, "Intrrupt for slot #%d is not clean\n", sc->slot_no[1]));
	sdshcbus_clear_intr(sc);
    }
/*
    SD_UNLOCK(sc);
*/
    return 0;
}


static int
sdshcbus_io_request(device_t dev,
		     u_int32_t cmd_index, u_int32_t arg,
		     int write, int blk_size, int num_blk, void *data)
{
    return sdshcbus_sd_transaction(device_get_softc(dev),
			      cmd_index, arg, write, blk_size, num_blk, data);
}


static int
sdshcbus_acmd_io_request(device_t dev,
		       u_int32_t acmd_index, u_int32_t arg,
		       int write, int blk_size, int num_blk, void *data)
{
    return sdshcbus_acmd_transaction(device_get_softc(dev),
				   acmd_index, arg, write, blk_size, num_blk,
				   data);
}


/****/

#define	SDSHCBUS_REG_READ1(sc, ofs)	\
	bus_space_read_1((sc)->bt, (sc)->bh, ofs)
#define	SDSHCBUS_REG_READ2(sc, ofs)	\
	bus_space_read_2((sc)->bt, (sc)->bh, ofs)
#define	SDSHCBUS_REG_READ4(sc, ofs)	\
	bus_space_read_4((sc)->bt, (sc)->bh, ofs)
#define	SDSHCBUS_READ1(sc, name)		SDSHCBUS_REG_READ1(sc, SDSHC_HSR_ ## name)
#define	SDSHCBUS_READ2(sc, name)		SDSHCBUS_REG_READ2(sc, SDSHC_HSR_ ## name)
#define	SDSHCBUS_READ4(sc, name)		SDSHCBUS_REG_READ4(sc, SDSHC_HSR_ ## name)

#define	SDSHCBUS_REG_WRITE1(sc, ofs, val)	\
	bus_space_write_1((sc)->bt, (sc)->bh, ofs, val)
#define	SDSHCBUS_REG_WRITE2(sc, ofs, val)	\
	bus_space_write_2((sc)->bt, (sc)->bh, ofs, val)
#define	SDSHCBUS_REG_WRITE4(sc, ofs, val)	\
	bus_space_write_4((sc)->bt, (sc)->bh, ofs, val)
#define	SDSHCBUS_WRITE1(sc, name, val)	SDSHCBUS_REG_WRITE1(sc, SDSHC_HSR_ ## name, val)
#define	SDSHCBUS_WRITE2(sc, name, val)	SDSHCBUS_REG_WRITE2(sc, SDSHC_HSR_ ## name, val)
#define	SDSHCBUS_WRITE4(sc, name, val)	SDSHCBUS_REG_WRITE4(sc, SDSHC_HSR_ ## name, val)


static int
sdshcbus_init(struct sdshcbus_softc *sc)
{
    device_t dev = sc->dev;
    device_t pdev = device_get_parent(dev);
    struct sdshc_softc *psc = device_get_softc(pdev);
    int *mem_rid = &psc->mem_rid[sc->slot_no[0]];
    int error;

    sc->slot_no[1] = sc->slot_no[0] + psc->slot_base;

    if (sc->slot_no[0] == 0) {
	sc->mem = psc->mem;
	sc->bt = psc->bt;
	sc->bh = psc->bh;
    } else {
	sc->mem = bus_alloc_resource(dev, SYS_RES_MEMORY, mem_rid,
				     0ul, ~0ul, 256, RF_ACTIVE);
	if (sc->mem == NULL) {
	    MSG((dev, "could not allocate memory resource\n"));
	    return ENOMEM;
	}
	sc->bt = rman_get_bustag(sc->mem);
	sc->bh = rman_get_bushandle(sc->mem);
    }

    sc->status = SDSHCBUS_STAT_NONE;

    if ((error = sdshcbus_config(sc)) != 0) {
	(void)sdshcbus_reset(sc);
	if ((error = sdshcbus_config(sc)) != 0) {
	    goto cleanup;
	}
    }

cleanup:
    return error;
}


static int
sdshcbus_config(struct sdshcbus_softc *sc)
{
    u_int32_t reg, tmp;

    LOG(0x10, (sc->dev, "Slot #%d\n", sc->slot_no[1]));
    reg = SDSHCBUS_READ4(sc, CAPABILITIES);
    sc->support_voltage =	(reg >> 24) & 0x07;
    sc->suspend_capable =	(reg >> 23) & 0x01;
    sc->dma_capable =		(reg >> 22) & 0x01;
    sc->highspeed_capable =	(reg >> 21) & 0x01;
    MSG((sc->dev, "capability:%s%s%s%s%s%s\n",
	 (sc->support_voltage & 0x01) ? " 3.3V" : "",
	 (sc->support_voltage & 0x02) ? " 3.0V" : "",
	 (sc->support_voltage & 0x04) ? " 1.8V" : "",
	 sc->suspend_capable ? " Suspend/Resume" : "",
	 sc->dma_capable ? " DMA" : "",
	 sc->highspeed_capable ? " High-Speed" : ""));
    LOG(0x10, (sc->dev, "  Voltage Support 1.8V: %s\n",
	       (sc->support_voltage & 0x04) ? "yes" : "no"));
    LOG(0x10, (sc->dev, "  Voltage Support 3.0V: %s\n",
	       (sc->support_voltage & 0x02) ? "yes" : "no"));
    LOG(0x10, (sc->dev, "  Voltage Support 3.3V: %s\n",
	       (sc->support_voltage & 0x01) ? "yes" : "no"));
    LOG(0x10, (sc->dev, "  Suspend/Resume Support: %s\n",
	       sc->suspend_capable ? "yes" : "no"));
    LOG(0x10, (sc->dev, "  DMA Support: %s\n",
	       sc->dma_capable ? "yes" : "no"));
    LOG(0x10, (sc->dev, "  High Speed Support: %s\n",
	       sc->highspeed_capable ? "yes" : "no"));
    tmp = (reg >> 16) & 0x03;
    sc->max_block_size = (tmp == 0x00 ?  512 :
			  tmp == 0x01 ? 1024 :
			  tmp == 0x02 ? 2048 :
			  0);
    if (sc->max_block_size == 0) {
	MSG((sc->dev,
	     "slot #%d: Max Block Length is Unknown\n", sc->slot_no[1]));
	return ENXIO;
    } else {
	LOG(0x10, (sc->dev, "  Max Block Length: %dB\n", sc->max_block_size));
    }
    sc->base_freq = (reg >> 8) & 0x3f;
    if (sc->base_freq == 0) {
	MSG((sc->dev, "slot #%d: Base Clock Freq for SD Clock is Unknown\n",
	     sc->slot_no[1]));
	return ENXIO;
    } else {
	LOG(0x10, (sc->dev, "  Base Clock Freq for SD Clock: %dMHz\n",
		   sc->base_freq));
    }
    sc->timeout_clock_freq = reg & 0x3f;
    tmp = (reg >> 7) & 0x01;
    if (sc->timeout_clock_freq == 0) {
	MSG((sc->dev,
	     "slot #%d: Timeout Clock Freq is Unknown\n", sc->slot_no[1]));
	return ENXIO;
    } else {
	LOG(0x10, (sc->dev, "  Timeout Clock Freq: %d%s\n",
		   sc->timeout_clock_freq, tmp ? "MHz" : "KHz"));
    }
    if (tmp) {
	sc->timeout_clock_freq *= 1000;
    }

#if defined(SD_DEBUG)
    if (0x10 <= sdshc_debug_level) {
	reg = SDSHCBUS_READ4(sc, MAXCURR_CAP);
	if ((tmp = (reg >> 16) & 0x0ff) == 0) {
	    MSG((sc->dev, "  Maximum Current for 1.8V: N/A\n"));
	} else {
	    MSG((sc->dev, "  Maximum Current for 1.8V: %dmA (0x%02x)\n",
		 tmp << 2, tmp));
	}
	if ((tmp = (reg >>  8) & 0x0ff) == 0) {
	    MSG((sc->dev, "  Maximum Current for 3.0V: N/A\n"));
	} else {
	    MSG((sc->dev, "  Maximum Current for 3.0V: %dmA (0x%02x)\n",
		 tmp << 2, tmp));
	}
	if ((tmp = (reg >>  0) & 0x0ff) == 0) {
	    MSG((sc->dev, "  Maximum Current for 3.3V: N/A\n"));
	} else {
	    MSG((sc->dev, "  Maximum Current for 3.3V: %dmA (0x%02x)\n",
		 tmp << 2, tmp));
	}
    }
#endif
    sc->card_inserted = sdshcbus_check_card_inserted(sc);
    LOG(0x10, (sc->dev, "  Card inserted: %s\n",
	       sc->card_inserted ? "yes" : "no"));

    sdshcbus_set_data_timeout(sc, 100);
    LOG(0x10, (sc->dev, "Enable Intr for slot #%d\n", sc->slot_no[1]));
    sc->flag.in_error_recovery = 0;
    sc->flag.error_recovery_wo_abort = 0;
    sc->flag.use_polling = 0;
    sdshcbus_clear_intr(sc);
    sdshcbus_enable_intr(sc);

    if (sc->card_inserted) {
	taskqueue_enqueue_fast(taskqueue_fast, &sc->attach_task);
    }

    LOG(0x08, (sc->dev, "Initialization successed\n"));
    return 0;
}


static int
ipow(int base, int n)
{
    int cnt, val;

    val = 1;
    for (cnt = 0; cnt < n; cnt++) {
	val *= base;
    }
    return val;
}


static void
sdshcbus_clear_intr(struct sdshcbus_softc *sc)
{
    u_int32_t reg;

    SDSHCBUS_WRITE2(sc, NORM_INTR_ST, 0x00ff);
    SDSHCBUS_WRITE2(sc, ERR_INTR_ST, 0xf1ff);
    reg = SDSHCBUS_READ2(sc, NORM_INTR_EN);
    reg &= ~0x0100;
    SDSHCBUS_WRITE2(sc, NORM_INTR_EN, reg);
    LOG(0x10, (sc->dev,
	       "Interrupt cleared for slot #%d\n", sc->slot_no[1]));
}


static void
sdshcbus_disable_intr(struct sdshcbus_softc *sc)
{
    u_int32_t reg;

    reg = SDSHCBUS_READ2(sc, NORM_INTR_EN);
    reg &= ~0x0100;
    SDSHCBUS_WRITE2(sc, NORM_INTR_EN, reg);

    SDSHCBUS_WRITE2(sc, NORM_INTR_SIGEN, 0x0000);
    SDSHCBUS_WRITE2(sc, ERR_INTR_SIGEN, 0x0000);
    LOG(0x10, (sc->dev,
	       "Interrupt disabled for slot #%d\n", sc->slot_no[1]));
}


static void
sdshcbus_enable_intr(struct sdshcbus_softc *sc)
{
    SDSHCBUS_WRITE2(sc, NORM_INTR_EN, 0x01ff);
    SDSHCBUS_WRITE2(sc, ERR_INTR_EN, 0xf1ff);
    SDSHCBUS_WRITE2(sc, NORM_INTR_SIGEN, 0x01ff);
    SDSHCBUS_WRITE2(sc, ERR_INTR_SIGEN, 0xf1ff);
    LOG(0x10, (sc->dev,
	       "Interrupt enabled for slot #%d\n", sc->slot_no[1]));
}


static void
sdshcbus_enable_intr_polling(struct sdshcbus_softc *sc)
{
    SDSHCBUS_WRITE2(sc, NORM_INTR_EN, 0x01ff);
    SDSHCBUS_WRITE2(sc, ERR_INTR_EN, 0xf1ff);
    SDSHCBUS_WRITE2(sc, NORM_INTR_SIGEN, 0x0000);
    SDSHCBUS_WRITE2(sc, ERR_INTR_SIGEN, 0x0000);
    LOG(0x10, (sc->dev,
	       "Interrupt enabled for slot #%d (polling)\n",
	       sc->slot_no[1]));
}


static void
sdshcbus_intr_handler(struct sdshcbus_softc *sc)
{
    u_int32_t reg;
    int card_detect = 0, no_wakeup = 0;

    sc->result = reg = SDSHCBUS_READ2(sc, NORM_INTR_ST);
    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
    
    LOG(0x08, (sc->dev, "Intr status of slot #%d: 0x%04x\n",
	       sc->slot_no[1], reg));
    if (reg & SDSHC_HSR_NORM_INTR_CARD_DETECT) {
	card_detect = 1;
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_CARD_DETECT);
    }

    if (reg & SDSHC_HSR_NORM_INTR_BUFRD_READY) {
	LOG(0x08, (sc->dev, "Buffer Read Ready Interrupt occured\n"));
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_BUFRD_READY);

	if (0 < sc->rest_blk_count && sc->addr != NULL) {
	    int cnt;

	    no_wakeup = 1;

	    /* `Get Block Data' */
	    for (cnt = sc->block_size; 4 <= cnt; cnt -= 4, sc->addr += 4) {
		*((u_int32_t *)sc->addr) = SDSHCBUS_READ4(sc, BUFDATA_PORT);
		LOG(0x20, (sc->dev, "read data: 0x%08x to %p\n",
			   *((u_int32_t *)sc->addr),
			   sc->addr));
	    }
	    if (0 < cnt) {
		u_int32_t reg;

		reg = SDSHCBUS_READ4(sc, BUFDATA_PORT);

		if (2 <= cnt) {
		    *((u_int16_t *)sc->addr) = (u_int16_t)reg;
		    LOG(0x20, (sc->dev, "read data: 0x%04x to %p\n",
			       (reg & 0xffff), sc->addr));
		    sc->addr += 2;
		    cnt -= 2;
		    reg >>= 16;
		}
		if (1 <= cnt) {
		    *sc->addr++ = (u_int8_t)reg;
		    LOG(0x20, (sc->dev, "read data: 0x%02x to %p\n",
			       (reg & 0xff), sc->addr));
		}
	    }
	    --sc->rest_blk_count;
	}
    }
    if (reg & SDSHC_HSR_NORM_INTR_BUFWR_READY) {
	LOG(0x08, (sc->dev, "Buffer Write Ready Interrupt occured\n"));
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_BUFWR_READY);

	if (0 < sc->rest_blk_count && sc->addr != NULL) {
	    int cnt;

	    no_wakeup = 1;

	    /* `Set Block Data' */
	    for (cnt = sc->block_size; 0 < cnt; cnt -= 4, sc->addr += 4) {
		LOG(0x20, (sc->dev, "write data: 0x%08x at %p\n",
			   *((u_int32_t *)sc->addr), sc->addr));
		SDSHCBUS_WRITE4(sc, BUFDATA_PORT, *((u_int32_t *)sc->addr));
	    }
	    if (cnt < 0) {
		sc->addr += cnt;
	    }
	    --sc->rest_blk_count;
	    LOG(0x08, (sc->dev, "write block remains: %d\n",
		       sc->rest_blk_count));
	}
    }
    if (reg & SDSHC_HSR_NORM_INTR_DMA_INTR) {
	LOG(0x08, (sc->dev, "DMA Interrupt occured\n"));
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_DMA_INTR);
    }
    if (reg & SDSHC_HSR_NORM_INTR_BLKGAP_EVENT) {
	LOG(0x08, (sc->dev, "Block Gap Event Interrupt occured\n"));
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_BLKGAP_EVENT);
    }
    if (reg & SDSHC_HSR_NORM_INTR_TRANS_COMPLETE) {
	LOG(0x08, (sc->dev, "Transfer Complete Interrupt occured\n"));
	sc->status |= SDSHCBUS_STAT_TRANSFER_COMPLETE;
	LOG(0x10, (sc->dev, "clear Transfer Complete Interrupt\n"));
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_TRANS_COMPLETE);
    }
    if (reg & SDSHC_HSR_NORM_INTR_CMD_COMPLETE) {
	LOG(0x08, (sc->dev, "Command Interrupt occured\n"));
	sc->status |= SDSHCBUS_STAT_COMMAND_COMPLETE;
	LOG(0x10, (sc->dev, "clear Command Interrupt\n"));
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_CMD_COMPLETE);
    }

    if (reg & SDSHC_HSR_NORM_INTR_ERR_INT) {
	MSG((sc->dev, "Err Intr status: 0x%04x\n", sc->error));
	sdshcbus_error_recovery(sc);
    } else {
	/* no error */
	sc->error = 0;
    }

    if (card_detect) {
	int card_state = sdshcbus_check_card_inserted(sc);
	int prev_state = (sc->card_inserted != SDDEV_NONE);

	LOG(0x10, (sc->dev, "p: %d, c: %d, i: %d\n",
		   prev_state, card_state, sc->card_inserted));
	if (prev_state == card_state) {
	    LOG(0x08, (sc->dev, "Card Detect Interrupt occured,"
		       " but status is not changed; ignored\n"));
	} else {
	    if (card_state) {
		tsleep(sc, PPAUSE, "sdshcbus_event_card_insertion",
		       sdshc_wait_after_insertion);
		if ((card_state = sdshcbus_check_card_inserted(sc))) {
		    LOG(0x08, (sc->dev, "Card Insertion detected\n"));
		    taskqueue_enqueue_fast(taskqueue_fast, &sc->attach_task);
		}
	    } else {
		LOG(0x08, (sc->dev, "Card Removal detected\n"));
		taskqueue_enqueue_fast(taskqueue_fast, &sc->detach_task);
	    }
	}
	SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_CARD_DETECT);
    } else {
	if (!no_wakeup) {
	    wakeup(sc);
	}
    }
}


static void
sdshcbus_error_recovery(struct sdshcbus_softc *sc)
{
    u_int16_t reg = sc->error;
    int cnt;

    LOG(0x08, (sc->dev, "Err Intr status: 0x%04x\n", reg));
    if (reg & SDSHC_HSR_ERR_INTR_CURRENT_LIMIT) {
	LOG(0x08, (sc->dev, "Current Limit Error Interrupt occured\n"));
	SDSHCBUS_WRITE2(sc, ERR_INTR_ST, SDSHC_HSR_ERR_INTR_CURRENT_LIMIT);
	MSG((sc->dev, "Current Limit Violation occurred\n"));
	goto fatal;
    }

    sc->flag.in_error_recovery = 1;

    if (reg & (SDSHC_HSR_ERR_INTR_CMD_ERROR | SDSHC_HSR_ERR_INTR_DATA_ERROR)) {
	if (reg & SDSHC_HSR_ERR_INTR_CMD_ERROR) {
	    if (reg & SDSHC_HSR_ERR_INTR_CMD_TIMEOUT) {
		LOG(0x08,
		    (sc->dev, "Command Timeout Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc,
			   ERR_INTR_ST, SDSHC_HSR_ERR_INTR_CMD_TIMEOUT);
	    }
	    if (reg & SDSHC_HSR_ERR_INTR_CMD_CRC) {
		LOG(0x08,
		    (sc->dev, "Command CRC Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc, ERR_INTR_ST, SDSHC_HSR_ERR_INTR_CMD_CRC);
	    }
	    if (reg & SDSHC_HSR_ERR_INTR_CMD_END_BIT) {
		LOG(0x08,
		    (sc->dev, "Command End Bit Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc,
			   ERR_INTR_ST, SDSHC_HSR_ERR_INTR_CMD_END_BIT);
	    }
	    if (reg & SDSHC_HSR_ERR_INTR_CMD_INDEX) {
		LOG(0x08,
		    (sc->dev, "Command Index Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc,
			   ERR_INTR_ST, SDSHC_HSR_ERR_INTR_CMD_INDEX);
	    }

	    if (sdshcbus_reset_cmd_line(sc)) {
		goto fatal;
	    }
	}

	if (reg & SDSHC_HSR_ERR_INTR_DATA_ERROR) {
	    if (reg & SDSHC_HSR_ERR_INTR_DATA_TIMEOUT) {
		LOG(0x08,
		    (sc->dev, "Data Timeout Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc,
			   ERR_INTR_ST, SDSHC_HSR_ERR_INTR_DATA_TIMEOUT);
	    }
	    if (reg & SDSHC_HSR_ERR_INTR_DATA_CRC) {
		LOG(0x08,
		    (sc->dev, "Data CRC Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc,
			   ERR_INTR_ST, SDSHC_HSR_ERR_INTR_DATA_CRC);
	    }
	    if (reg & SDSHC_HSR_ERR_INTR_DATA_END_BIT) {
		LOG(0x08,
		    (sc->dev, "Data End Bit Error Interrupt occured\n"));
		SDSHCBUS_WRITE2(sc,
			   ERR_INTR_ST, SDSHC_HSR_ERR_INTR_DATA_END_BIT);
	    }

	    if (sdshcbus_reset_dat_line(sc)) {
		goto fatal;
	    }
	}

	if (!sc->flag.error_recovery_wo_abort) {
	    sdshcbus_enable_intr_polling(sc);
	    if (sdshcbus_async_abort(sc)) {
		goto fatal;
	    }

	    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		if ((SDSHCBUS_READ4(sc, PRESENT_STATE)
		     & SDSHC_HSR_PRESENT_STATE_CMD_BUSY) == 0) {
		    break;
		}
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
		SD_SLEEP(sc, PPAUSE,
			 "sdshcbus_error_recovery", sdshc_polling_interval);
	    }
	    if (SDSHC_TIMEOUT_COUNT <= cnt) {
		MSG((sc->dev, "Command Inhibit clear timeout; giving up\n"));
		goto fatal;
	    }

	    if (SDSHCBUS_READ2(sc, ERR_INTR_ST) & SDSHC_HSR_ERR_INTR_CMD_ERROR) {
		goto fatal;
	    }
	    if (SDSHCBUS_READ2(sc, ERR_INTR_ST)
					& SDSHC_HSR_ERR_INTR_DATA_TIMEOUT) {
		goto fatal;
	    }

	    tsleep(sc, PPAUSE, "sdshcbus_error_recovery", hz / 100);

	    if (sc->flag.use_polling) {
		sdshcbus_disable_intr(sc);
	    }
	}

	if ((SDSHCBUS_READ4(sc, PRESENT_STATE) & SDSHC_HSR_PRESENT_STATE_DATLINE)
					    != SDSHC_HSR_PRESENT_STATE_DATLINE) {
	    goto fatal;
	}

	/* revocerable error */
    }

    if (reg & SDSHC_HSR_ERR_INTR_AUTO_CMD12) {
	LOG(0x08, (sc->dev, "Auto CMD12 Error Interrupt occured\n"));
	SDSHCBUS_WRITE2(sc, ERR_INTR_ST, SDSHC_HSR_ERR_INTR_AUTO_CMD12);
    }

    if (reg & SDSHC_HSR_ERR_INTR_VENDOR) {
	LOG(0x08, (sc->dev, "Vendor Specific Error Interrupt occured (0x%x)\n",
		   (reg & SDSHC_HSR_ERR_INTR_VENDOR) >> 12));
	SDSHCBUS_WRITE2(sc, ERR_INTR_ST, SDSHC_HSR_ERR_INTR_VENDOR);
    }

    reg = SDSHCBUS_READ2(sc, ERR_INTR_ST);
    LOG(0x10, (sc->dev, "Err Intr status after recovery: 0x%04x\n", reg));

    sc->flag.in_error_recovery = 0;
    return;

fatal:
    /* non-recoverable error */
    MSG((sc->dev, "Fatal error occurred at slot #%d; power off for safety\n",
	 sc->slot_no[1]));
    sdshcbus_card_powerdown(sc);
    sc->card_inserted = SDDEV_INVALID;
    sc->flag.in_error_recovery = 0;
}


static int
sdshcbus_check_intr_clean(struct sdshcbus_softc *sc)
{
    u_int32_t reg, tmp;

/*
    LOG(0x20, (sc->dev, "Normal Interrupt Status: 0x%04x\n",
	       SDSHCBUS_READ2(sc, NORM_INTR_ST)));
    LOG(0x20, (sc->dev, "Error Interrupt Status: 0x%04x\n",
	       SDSHCBUS_READ2(sc, ERR_INTR_ST)));
    LOG(0x20, (sc->dev, "Normal Interrupt Enable: 0x%04x\n",
	       SDSHCBUS_READ2(sc, NORM_INTR_EN)));
    LOG(0x20, (sc->dev, "Error Interrupt Enable: 0x%04x\n",
	       SDSHCBUS_READ2(sc, ERR_INTR_EN)));
    LOG(0x20, (sc->dev, "Normal Interrupt Signal Enable: 0x%04x\n",
	       SDSHCBUS_READ2(sc, NORM_INTR_SIGEN)));
    LOG(0x20, (sc->dev, "Error Interrupt Signal Enable: 0x%04x\n",
	       SDSHCBUS_READ2(sc, ERR_INTR_SIGEN)));
*/

    reg = SDSHCBUS_READ2(sc, NORM_INTR_ST);
    tmp = reg & (SDSHC_HSR_NORM_INTR_ERR_INT | SDSHC_HSR_NORM_INTR_CARD_INT);
    if (tmp != 0) {
	LOG(0x08, (sc->dev, "Intrrupt status is not clean (0x%04x)\n", reg));
    }
    return !tmp;
}




static int
sdshcbus_reset_lines(struct sdshcbus_softc *sc)
{
    int cnt;

    LOG(0x08, (sc->dev, "Resetting CMD / DAT line...\n"));
    SDSHCBUS_WRITE1(sc, RESET, SDSHC_HSR_RESET_LINE);
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ1(sc, RESET) & SDSHC_HSR_RESET_LINE) == 0) {
	    return 0;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_reset_lines", sdshc_polling_interval);
    }
    MSG((sc->dev, "Line reset failure\n"));
    return ENXIO;
}


static int
sdshcbus_reset_cmd_line(struct sdshcbus_softc *sc)
{
    int cnt;

    LOG(0x08, (sc->dev, "Resetting CMD line...\n"));
    SDSHCBUS_WRITE1(sc, RESET, SDSHC_HSR_RESET_CMD);
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ1(sc, RESET) & SDSHC_HSR_RESET_CMD) == 0) {
	    return 0;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_reset_cmd_lines", sdshc_polling_interval);
    }
    MSG((sc->dev, "CMD line reset failure\n"));
    return ENXIO;
}


static int
sdshcbus_reset_dat_line(struct sdshcbus_softc *sc)
{
    int cnt;

    LOG(0x08, (sc->dev, "Resetting DAT line...\n"));
    SDSHCBUS_WRITE1(sc, RESET, SDSHC_HSR_RESET_DAT);
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ1(sc, RESET) & SDSHC_HSR_RESET_DAT) == 0) {
	    return 0;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_reset_dat_lines", sdshc_polling_interval);
    }
    MSG((sc->dev, "DAT line reset failure\n"));
    return ENXIO;
}


static int
sdshcbus_reset(struct sdshcbus_softc *sc)
{
    int cnt;

    LOG(0x08, (sc->dev, "Resetting device...\n"));
    SDSHCBUS_WRITE1(sc, RESET, SDSHC_HSR_RESET_ALL);
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ1(sc, RESET) & SDSHC_HSR_RESET_MASK) == 0) {
	    return 0;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_reset", sdshc_polling_interval);
    }
    MSG((sc->dev, "Controller reset failure\n"));
    return ENXIO;
}


static void
sdshcbus_stop(struct sdshcbus_softc *sc)
{
/*    SD_LOCK(sc); */
    sdshcbus_disable_intr(sc);
    sdshcbus_card_powerdown(sc);
    sc->card_inserted = SDDEV_INVALID;
/*    SD_UNLOCK(sc); */
}


static void
sdshcbus_card_insertion_callback(void *arg, int count)
{
    struct sdshcbus_softc *sc = (struct sdshcbus_softc *)arg;

/*
    printf( "Card insertion callback invoked\n");
    printf( "Card insertion callback invoked; sc = %p\n", sc);
    printf( "Card insertion callback invoked; sc = %p, *sc = %p\n",
	    sc, *sc);
*/

    LOG(0x10, (sc->dev, "Card insertion callback invoked at slot #%d\n",
	       sc->slot_no[1]));
/*    SD_LOCK(sc); - DO NOT LOCK HERE */
    sdshcbus_event_card_insertion(sc);
/*    SD_UNLOCK(sc); */
}


static void
sdshcbus_card_removal_callback(void *arg, int count)
{
    struct sdshcbus_softc *sc = (struct sdshcbus_softc *)arg;

    LOG(0x10, (sc->dev, "Card removal callback invoked at slot #%d\n",
	       sc->slot_no[1]));
/*    SD_LOCK(sc); - DO NOT LOCK HERE */
    sdshcbus_event_card_removal(sc);
/*    SD_UNLOCK(sc); */
}


static void
sdshcbus_event_card_insertion(struct sdshcbus_softc *sc)
{
    int card_state = sdshcbus_check_card_inserted(sc);
    int err;

/*
    tsleep(sc,
	   PPAUSE, "sdshcbus_event_card_insertion", sdshc_wait_after_insertion);
*/
    if (!(sc->card_inserted = card_state)) {
	LOG(0x08, (sc->dev, "card is not inserted\n"));
    }
    MSG((sc->dev, "card inserted in slot #%d\n", sc->slot_no[1]));

    sdshcbus_disable_intr(sc);
    sc->flag.use_polling = 1;
    err = sdshcbus_card_init(sc);
    sc->flag.use_polling = 0;
    sdshcbus_enable_intr(sc);

    if (!err) {
	sdshcbus_led_control(sc, 0);

	device_add_child(sc->dev, "sd", sc->slot_no[0]);
	bus_generic_attach(sc->dev);
    }
}


static void
sdshcbus_event_card_removal(struct sdshcbus_softc *sc)
{
    MSG((sc->dev, "card removed\n"));

    sdshcbus_detach_children(sc->dev);

    sdshcbus_led_control(sc, 0);
    sdshcbus_clock_stop(sc);
    sdshcbus_card_powerdown(sc);
    sc->card_inserted = SDDEV_NONE;
}


static int
sdshcbus_check_card_inserted(struct sdshcbus_softc *sc)
{
    return ((SDSHCBUS_READ4(sc, PRESENT_STATE)
	     & SDSHC_HSR_PRESENT_STATE_CARD_INSERTED) != 0);
}


static int
sdshcbus_check_command_running(struct sdshcbus_softc *sc)
{
    return (SDSHCBUS_READ4(sc, PRESENT_STATE)
	    & SDSHC_HSR_PRESENT_STATE_CMD_BUSY);
}


static int
sdshcbus_check_clock_supplied(struct sdshcbus_softc *sc)
{
    return (SDSHCBUS_READ2(sc, CLOCK_CTL) & 0x07) == 0x07;
}


static void
sdshcbus_print_card_type(struct sdshcbus_softc *sc)
{
    char *card_type;

    switch (sc->card_inserted) {
      case SDDEV_INVALID:
	card_type = "invalid";
	break;
      case SDDEV_SD1:
	card_type = "SD card (Ver 1.xx compliant)";
	break;
      case SDDEV_SD2:
	card_type = "SD card (Ver 2.xx or later compliant)";
	break;
      case SDDEV_SDHC:
	card_type = "SDHC card";
	break;
      case SDDEV_SDIO:
	card_type = "SDIO";
	break;
      case SDDEV_MMC:
	card_type = "MMC";
	break;
      default:
	card_type = "Unknown";
	break;
    }
    LOG(0x08, (sc->dev,
	       "Card type: %s; capacity = %llu bytes, max clock = %ukHz\n",
	       card_type,
	       (u_int64_t)sc->card_capacity << 9, sc->card_max_clock));
}


static int
sdshcbus_led_control(struct sdshcbus_softc *sc, int state)
{
    u_int32_t reg;
    int prev;

    reg = SDSHCBUS_READ1(sc, HOST_CTL);
    prev = reg & SDSHC_HSR_HOST_CTL_LED;
    reg &= ~SDSHC_HSR_HOST_CTL_LED;
    if (state < 0) {
	reg ^= SDSHC_HSR_HOST_CTL_LED;
    } else {
	reg |= (state & 0x01);
    }
    SDSHCBUS_WRITE1(sc, HOST_CTL, reg);

    return prev;
}


static int
sdshcbus_clock_supply(struct sdshcbus_softc *sc, int clock)
{
    /* Unit of Clock frequency is KHz */
    int base = sc->base_freq * 1000;
    int divisor;
    u_int16_t reg;
    int cnt;

    /* `Calculate a divisor for SD clock frequency' */
    for (divisor = 1; divisor < 0x100; divisor <<= 1) {
	if (base / divisor <= clock) {
	    break;
	}
    }
    sc->card_clock = base / divisor;
    LOG(0x08, (sc->dev,
	       "target clock is %dKHz, divisor = %d, supply = %dkHz\n",
	       clock, divisor, sc->card_clock));
    divisor >>= 1;

    /* `Set SDCLK Frequency Select & Internal Clock Enable' */
    reg = (divisor << 8) | SDSHC_HSR_CLOCK_CTL_INTERNAL_CLK_EN;
    SDSHCBUS_WRITE2(sc, CLOCK_CTL, reg);

    /* wait Internal Clock becomes stable */
    LOG(0x08, (sc->dev, "wait clock becoming stable...\n"));
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ2(sc, CLOCK_CTL)
	     & SDSHC_HSR_CLOCK_CTL_INTERNAL_CLK_STABLE) != 0) {
	    break;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_clock_supply", sdshc_polling_interval);
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Controller Internal Clock is not stable; giving up\n"));
	sc->card_clock = 0;
	return 1;
    }

    /* `Set SD Clock On' */
    reg |= SDSHC_HSR_CLOCK_CTL_SD_CLK_EN;
    SDSHCBUS_WRITE2(sc, CLOCK_CTL, reg);
    LOG(0x08, (sc->dev, "clock supplied\n"));

    return 0;
}


static int
sdshcbus_clock_stop(struct sdshcbus_softc *sc)
{
    u_int16_t reg;

    if (sdshcbus_check_command_running(sc)) {
	return 1;
    }

    /* `Set SD Clock Off' */
    reg = SDSHCBUS_READ2(sc, CLOCK_CTL);
    reg &= ~SDSHC_HSR_CLOCK_CTL_SD_CLK_EN;
    SDSHCBUS_WRITE2(sc, CLOCK_CTL, reg);

    return 0;
}


static int
sdshcbus_clock_frequency_change(struct sdshcbus_softc *sc, int clock)
{
    if (sdshcbus_check_clock_supplied(sc)) {
	LOG(0x08, (sc->dev, "Stop SD Bus clock\n"));
	if (sdshcbus_clock_stop(sc)) {
	    return 1;
	}
    }

    SD_SLEEP(sc, PPAUSE, "sdshcbus_frequency_change", 1);

    LOG(0x08, (sc->dev, "frequency change to %dkHz\n", clock));
    return sdshcbus_clock_supply(sc, clock);
}


static void
sdshcbus_power_on(struct sdshcbus_softc *sc)
{
    u_int8_t reg;

    reg = SDSHCBUS_READ1(sc, POWER_CTL);
    reg |= 0x01;
    SDSHCBUS_WRITE1(sc, POWER_CTL, reg);
}


static void
sdshcbus_card_powerup(struct sdshcbus_softc *sc)
{
    u_int32_t reg;

    if (sc->card_voltage == 0) {
	/* card voltage not detected */

	/* Get the support voltage of the Host Controller */
	reg = sc->support_voltage;

	/* Set SD Bus Voltage Select with supported maximum voltage */
	if (reg & 0x01) {
	    reg = 0x0e;	/* 3.3V */
	    sc->card_voltage = 33;
	} else if ((reg >>= 1) & 0x01) {
	    reg = 0x0c;	/* 3.0V */
	    sc->card_voltage = 30;
	} else {
	    reg = 0x0a;	/* 1.8V */
	    sc->card_voltage = 18;
	}
    } else {
	/* card voltage detected */

	if (sc->card_voltage == 33) {
	    reg = 0x0e;	/* 3.3V */
	} else if (sc->card_voltage == 30) {
	    reg = 0x0c;	/* 3.0V */
	} else {
	    reg = 0x0a;	/* 1.8V */
	}
    }
    /* `Set SD Bus Voltage Select' */
    LOG(0x10, (sc->dev, "SD Bus Voltage selected to %d.%dV\n",
	       sc->card_voltage / 10,
	       sc->card_voltage % 10));
    SDSHCBUS_WRITE1(sc, POWER_CTL, reg);

    /* `Set SD Bus Power' */
    LOG(0x10, (sc->dev, "SD Bus power up\n"));
    reg |= 0x01;
    SDSHCBUS_WRITE1(sc, POWER_CTL, reg);
}


static void
sdshcbus_card_powerdown(struct sdshcbus_softc *sc)
{
    u_int8_t reg;

    reg = SDSHCBUS_READ1(sc, POWER_CTL);
    reg &= ~0x01;
    SDSHCBUS_WRITE1(sc, POWER_CTL, reg);
}


static int
sdshcbus_card_bootstrap(struct sdshcbus_softc *sc)
{
    /* `Get the support voltage of the Host Controller' */
    /* `Set SD Bus Voltage Select with supported maximum voltage' */
    /* `Set SD Bus Power' */
    LOG(0x10, (sc->dev,
	       "Setting up SD Bus Voltage with supported maximum voltage\n"));
    sc->card_voltage = 0;
    sc->card_max_clock = 0;
    sc->rca = 0;
    sdshcbus_card_powerup(sc);

    sdshcbus_set_highspeed_mode(sc, 0);
    sc->highspeed_enabled = 0;

    sdshcbus_clock_supply(sc, SDSHC_INITIAL_CLOCK);

    /* Issue CMD0 */
    if (sdshcbus_sd_transaction(sc, 0, 0, 0, 0, 0, NULL)) {
	MSG((sc->dev, "Timeout occured for CMD0; giving up\n"));
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	sc->card_inserted = SDDEV_INVALID;
	return 1;
    }

    return 0;
}


static int time_value[16] = {
    0, 10, 12, 13, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 70, 80
};

static int
sdshcbus_card_init(struct sdshcbus_softc *sc)
{
    u_int32_t reg, reg2;
    char *ptr;
    int voltage = sc->card_voltage;
    int err;
    u_int8_t scrbuf[8];
    u_int8_t sd_status[64];
    u_int8_t funcstat[64];
    u_int32_t tmp;
    int ret;

    LOG(0x08, (sc->dev, "Card initialization started\n"));

    if ((err = sdshcbus_card_bootstrap(sc))) {
	return err;
    }
    LOG(0x10, (sc->dev, "card reset (CMD0) done\n"));

    sc->flag.error_recovery_wo_abort = 1;

    /*
     * `Get OCR Value of SD Card'
     */
    sc->card_inserted = SDDEV_UNKNOWN;
    sc->sd_spec = -1;

    if (sdshcbus_sd_transaction(sc, 8, 0x01aa, 0, 0, 0, NULL)) {
	/* Timeout - Ver 2.xx SD (Voltage mismatch) or Ver 1.xx SD or not SD */
	LOG(0x10, (sc->dev,
		   "CMD8 timeout - Ver 2.xx or later SD (Voltage mismatch)"
		   " or Ver 1.xx SD or not SD\n"));
	sc->card_inserted = SDDEV_UNKNOWN;

	LOG(0x10, (sc->dev, "Re-issue CMD0\n"));
	if (sdshcbus_sd_transaction(sc, 0, 0, 0, 0, 0, NULL)) {
	    return EIO;
	}
	sdshcbus_set_highspeed_mode(sc, 0);
	sc->highspeed_enabled = 0;
	sdshcbus_clock_supply(sc, SDSHC_INITIAL_CLOCK);
    } else {
	/* Got response - Ver 2.xx SD */
	sc->card_inserted = SDDEV_SD2;

	reg = SDSHCBUS_READ4(sc, RESPONSE0);
	if (reg != 0x01aa) {
	    MSG((sc->dev,
		 "Invalid response received for CMD8;"
		 " this card is unusable\n"));
	    sc->card_inserted = SDDEV_INVALID;
	    sdshcbus_reset_lines(sc);
	    sdshcbus_card_powerdown(sc);
	    return ENXIO;
	}
	/* Card with compatible voltage */
	LOG(0x10, (sc->dev,
		   "Got valid response from CMD8"
		   " - Ver 2.xx or later compliant SD"
		   " with compatible voltage\n"));
    }

    if (sdshcbus_acmd_transaction(sc, 41, 0, 0, 0, 0, NULL)) {
	/* Timeout? */
	if (sc->card_inserted == SDDEV_UNKNOWN) {
	    LOG(0x08, (sc->dev, "Timeout occurred for ACMD41; maybe MMC\n"));
	    return sdshcbus_card_init_mmc(sc);
	} else {
	    MSG((sc->dev, "Timeout occurred for ACMD41;"
		 " this card is unusable\n"));
	}
	sc->card_inserted = SDDEV_INVALID;
	LOG(0x10, (sc->dev, "power off\n"));
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	LOG(0x10, (sc->dev, "power on\n"));
	sdshcbus_power_on(sc);
	sdshcbus_set_highspeed_mode(sc, 0);
	sc->highspeed_enabled = 0;
	sdshcbus_clock_supply(sc, SDSHC_INITIAL_CLOCK);
	LOG(0x10, (sc->dev, "Re-issue CMD0\n"));
	sdshcbus_sd_transaction(sc, 0, 0, 0, 0, 0, NULL);
	return ENXIO;
    }

    reg = SDSHCBUS_READ4(sc, RESPONSE0);
    LOG(0x10, (sc->dev, "resp: 0x%08x\n", reg));

    /* check card voltage */
    if ((sc->support_voltage & 0x2) && (reg & 0x00020000)) {
	/* contoller support 3.0V, card support 2.9 - 3.0V */
	reg2 = 0x00020000;
	voltage = 30;
    } else if ((sc->support_voltage & 0x2) && (reg & 0x00040000)) {
	/* contoller support 3.0V, card support 3.0 - 3.1V */
	reg2 = 0x00040000;
	voltage = 30;
    } else if ((sc->support_voltage & 0x1) && (reg & 0x00100000)) {
	/* contoller support 3.3V, card support 3.2 - 3.3V */
	reg2 = 0x00100000;
	voltage = 33;
    } else if ((sc->support_voltage & 0x1) && (reg & 0x00200000)) {
	/* contoller support 3.3V, card support 3.3 - 3.4V */
	reg2 = 0x00200000;
	voltage = 33;
    } else {
	MSG((sc->dev, "Voltage miss-match; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
/*
    reg2 = reg;
*/
    if (sc->card_inserted == SDDEV_SD2) {
	reg2 |= 0x40000000;
    }
    LOG(0x10, (sc->dev, "arg of ACMD41: 0x%08x\n", reg2));
    while (!(reg & 0x80000000)) {
	if (sdshcbus_acmd_transaction(sc, 41, reg2, 0, 0, 0, NULL)) {
	    return ENXIO;
	}
	reg = SDSHCBUS_READ4(sc, RESPONSE0);
	LOG(0x10, (sc->dev, "resp: 0x%08x\n", reg));
	if ((reg & 0x40000000) && !(reg2 & 0x40000000)) {
	    reg2 |= 0x40000000;
	    LOG(0x10, (sc->dev, "arg of ACMD41 changed to 0x%08x\n", reg2));
	}
    }
    LOG(0x10, (sc->dev, "OCR = 0x%08x\n", reg));
    sc->ocr = reg;
    if (sc->card_inserted == SDDEV_SD2) {
	if (reg & 0x40000000) {
	    sc->card_inserted = SDDEV_SDHC;
	}
    } else {
	sc->card_inserted = SDDEV_SD1;
    }

    /* `SD Bus Voltage changed?' */
    if (voltage != sc->card_voltage) {
	MSG((sc->dev, "changing card voltage to %d.%dV\n",
	     voltage / 10, voltage % 10));
	/* `Clr SD Bus Power' */
	sdshcbus_card_powerdown(sc);
	/* `Set SD Bus Voltage Select' */
	sc->card_voltage = voltage;
	sdshcbus_card_powerup(sc);
    }

    /* Identation by CMD2 */
    if (sdshcbus_sd_transaction(sc, 2, 0, 0, 0, 0, NULL)) {
	/* Timeout? */
	MSG((sc->dev, "Timeout occurred for CMD2; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
#if defined(SD_DEBUG)
    if (1) {
	u_int32_t resp0, resp1, resp2, resp3;

	resp0 = SDSHCBUS_READ4(sc, RESPONSE0);
	resp1 = SDSHCBUS_READ4(sc, RESPONSE1);
	resp2 = SDSHCBUS_READ4(sc, RESPONSE2);
	resp3 = SDSHCBUS_READ4(sc, RESPONSE3);
	LOG(0x10, (sc->dev, "CID: 0x%08x:%08x:%08x:%08x\n",
		   resp3, resp2, resp1, resp0));
    }
#endif
    reg = SDSHCBUS_READ4(sc, RESPONSE3);
    sc->mid = (reg >> 16) & 0xff;
    ptr = sc->oid;
    if (isprint((*ptr = (reg >> 8) & 0xff))) {
	++ptr;
    }
    if (isprint((*ptr = reg & 0xff))) {
	++ptr;
    }
    *ptr = '\0';
    reg = SDSHCBUS_READ4(sc, RESPONSE2);
    sc->pnm[0] = (char)(reg >> 24);
    sc->pnm[1] = (char)(reg >> 16);
    sc->pnm[2] = (char)(reg >>  8);
    sc->pnm[3] = (char)(reg >>  0);
    reg = SDSHCBUS_READ4(sc, RESPONSE1);
    sc->pnm[4] = (char)(reg >> 24);
    sc->pnm[5] = '\0';
    sc->prv[0] = (reg >> 20) & 0x0f;
    sc->prv[1] = (reg >> 16) & 0x0f;
    reg2 = SDSHCBUS_READ4(sc, RESPONSE0);
    LOG(0x08, (sc->dev, "Idented: %s rev %d.%d S/N %u (%d/%02d)\n",
	       sc->pnm, sc->prv[0], sc->prv[1], (reg << 16) | (reg2 >> 16),
	       2000 + ((reg2 >> 4) & 0xff), reg2 & 0xf));

    /* get RCA number by CMD3 */
    if (sdshcbus_sd_transaction(sc, 3, 0, 0, 0, 0, NULL)) {
	/* Timeout? */
	MSG((sc->dev, "Timeout occurred for CMD3; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
    reg = SDSHCBUS_READ4(sc, RESPONSE0);
    sc->rca = reg & 0xffff0000;
    LOG(0x10, (sc->dev, "RCA: 0x%04x,"
	       " COM_CRC_ERROR: %d, ILLEGAL_COMMAND: %d, ERROR: %d,"
	       " CURRENT_STATE: %s, READY_FOR_DATA: %d, APP_CMD: %d,"
	       " AKE_SEQ_ERROR: %d\n",
	       sc->rca >> 16,
	       (reg >> 23) & 0x01, (reg >> 22) & 0x01, (reg >> 19) & 0x01,
	       card_state[(reg >> 9) & 0x0f], (reg >> 8) & 0x01,
	       (reg >> 5) & 0x01, (reg >> 3) & 0x01));

    /* get CSD by CMD9 */
    if ((ret = sdshcbus_sd_get_CSD(sc)) != 0) {
	return ret;
    }

    sdshcbus_print_card_type(sc);

    if (sc->card_max_clock != 0) {
	sdshcbus_clock_frequency_change(sc, sc->card_max_clock);
    }

    /* Addressing card */
    if (sdshcbus_sd_transaction(sc, 7, sc->rca, 0, 0, 0, NULL)) {
	MSG((sc->dev, "CMD7 (SELECT_CARD: 0x%08x) failed\n", sc->rca));
	return ENXIO;
    }

    if (sc->card_inserted != SDDEV_SDHC) {
	/* Set block size */
	int blk_size = sc->block_size;
	if (sdshcbus_sd_transaction(sc, 16, blk_size, 0, 0, 0, NULL)) {
	    MSG((sc->dev, "CMD16 (SET_BLOCKLEN: %d) failed\n", blk_size));
	    return ENXIO;
	}
    }

    /* get SCR */
    if (sdshcbus_acmd_transaction(sc, 51, 0, 0, 8, 1, scrbuf)) {
	MSG((sc->dev, "ACMD51 (SEND_SCR) failed\n"));
    } else {
	LOG(0x10, (sc->dev, "SCR: 0x%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
		   scrbuf[0], scrbuf[1], scrbuf[2], scrbuf[3],
		   scrbuf[4], scrbuf[5], scrbuf[6], scrbuf[7]));
	LOG(0x10, (sc->dev, "SCR_STRUCTURE: %s\n",
		   ((scrbuf[0] >> 4) & 0x0f) == 0 ? "SCR ver 1.0" :
		   "(reserved)"));
	sc->sd_spec = scrbuf[0] & 0x0f;
	LOG(0x10, (sc->dev, "SD_SPEC: %s\n",
		   sc->sd_spec == 0 ? "Ver 1.0-1.01" :
		   sc->sd_spec == 1 ? "Ver 1.10" :
		   sc->sd_spec == 2 ? "Ver 2.00" :
		   "(reserved)"));
	LOG(0x10, (sc->dev, "DATA_STAT_AFTER_ERASE: %d\n",
		   (scrbuf[1] >> 7) & 0x01));
	LOG(0x10, (sc->dev, "SD_SECURITY: %s\n",
		   ((scrbuf[1] >> 4) & 0x07) == 0 ? "no security" :
		   ((scrbuf[1] >> 4) & 0x07) == 1 ? "Not used" :
		   ((scrbuf[1] >> 4) & 0x07) == 2 ? "Ver 1.01" :
		   ((scrbuf[1] >> 4) & 0x07) == 3 ? "Ver 2.00" :
		   "(reserved)"));
	LOG(0x10, (sc->dev, "SD_BUS_WIDTHS: 0x%x\n",
		   scrbuf[1] & 0x0f));
	sc->card_max_buswidth = ((scrbuf[1] & 0x04) ? 4 :
				 (scrbuf[1] & 0x01) ? 1 :
				 0);
	if (sc->card_max_buswidth == 0) {
	    LOG(0x10, (sc->dev, "Couldn't get bus width; assuming 1-bit\n"));
	    sc->card_max_buswidth = 1;
	} else {
	    LOG(0x10, (sc->dev, "max bus width: %d-bit\n",
		       sc->card_max_buswidth));
	    sdshcbus_change_bus_width(sc, sc->card_max_buswidth);
	}
    }

    if (1 <= sc->sd_spec && sc->highspeed_capable) {
	/* If Ver 1.10 or later, check high speed mode supported by card */
	if (sdshcbus_sd_transaction(sc, 6, 0x00fffff1, 0, 64, 1, funcstat)) {
	    MSG((sc->dev, "CMD6 (SWITCH_FUNC, mode 0) failed\n"));
	} else {
	    LOG(0x10, (sc->dev, "Switch Function Status: 0x"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		      "%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
		      funcstat[ 0], funcstat[ 1], funcstat[ 2], funcstat[ 3],
		      funcstat[ 4], funcstat[ 5], funcstat[ 6], funcstat[ 7],
		      funcstat[ 8], funcstat[ 9], funcstat[10], funcstat[11],
		      funcstat[12], funcstat[13], funcstat[14], funcstat[15],
		      funcstat[16], funcstat[17], funcstat[18], funcstat[19],
		      funcstat[20], funcstat[21], funcstat[22], funcstat[23],
		      funcstat[24], funcstat[25], funcstat[26], funcstat[27],
		      funcstat[28], funcstat[29], funcstat[30], funcstat[31],
		      funcstat[32], funcstat[33], funcstat[34], funcstat[35],
		      funcstat[36], funcstat[37], funcstat[38], funcstat[39],
		      funcstat[40], funcstat[41], funcstat[42], funcstat[43],
		      funcstat[44], funcstat[45], funcstat[46], funcstat[47],
		      funcstat[48], funcstat[49], funcstat[50], funcstat[51],
		      funcstat[52], funcstat[53], funcstat[54], funcstat[55],
		      funcstat[56], funcstat[57], funcstat[58], funcstat[59],
		      funcstat[60], funcstat[61], funcstat[62], funcstat[63]));
	}
	LOG(0x10, (sc->dev, "Maximum Current Consumption: %dmA\n",
		   funcstat[ 0] << 8 | funcstat[ 1]));
	LOG(0x10, (sc->dev, "Function group 6 info: 0x%04x\n",
		   funcstat[ 2] << 8 | funcstat[ 3]));
	LOG(0x10, (sc->dev, "Function group 5 info: 0x%04x\n",
		   funcstat[ 4] << 8 | funcstat[ 5]));
	LOG(0x10, (sc->dev, "Function group 4 info: 0x%04x\n",
		   funcstat[ 6] << 8 | funcstat[ 7]));
	LOG(0x10, (sc->dev, "Function group 3 info: 0x%04x\n",
		   funcstat[ 8] << 8 | funcstat[ 9]));
	LOG(0x10, (sc->dev, "Function group 2 info: 0x%04x\n",
		   funcstat[10] << 8 | funcstat[11]));
	LOG(0x10, (sc->dev, "Function group 1 info: 0x%04x\n",
		   funcstat[12] << 8 | funcstat[13]));
	LOG(0x10,
	    (sc->dev, "The function which can be switched in group 6: 0x%x\n",
	     funcstat[14] >> 4));
	LOG(0x10,
	    (sc->dev, "The function which can be switched in group 5: 0x%x\n",
	     funcstat[14] & 0x0f));
	LOG(0x10,
	    (sc->dev, "The function which can be switched in group 4: 0x%x\n",
	     funcstat[15] >> 4));
	LOG(0x10,
	    (sc->dev, "The function which can be switched in group 3: 0x%x\n",
	     funcstat[15] & 0x0f));
	LOG(0x10,
	    (sc->dev, "The function which can be switched in group 2: 0x%x\n",
	     funcstat[16] >> 4));
	LOG(0x10,
	    (sc->dev, "The function which can be switched in group 1: 0x%x\n",
	     funcstat[16] & 0x0f));
	LOG(0x10, (sc->dev, "Data Structure Version: %s\n",
		   funcstat[17] == 0x00 ? "bits 511:376 are defined" :
		   funcstat[17] == 0x01 ? "bits 511:272 are defined" :
		   "(reserved)"));
	if (funcstat[17] == 0x01) {
	    LOG(0x10,
		(sc->dev, "Busy Status of functions in group 6: 0x%04x\n",
		 funcstat[18] << 8 | funcstat[19]));
	    LOG(0x10,
		(sc->dev, "Busy Status of functions in group 5: 0x%04x\n",
		 funcstat[20] << 8 | funcstat[21]));
	    LOG(0x10,
		(sc->dev, "Busy Status of functions in group 4: 0x%04x\n",
		 funcstat[22] << 8 | funcstat[23]));
	    LOG(0x10,
		(sc->dev, "Busy Status of functions in group 3: 0x%04x\n",
		 funcstat[24] << 8 | funcstat[25]));
	    LOG(0x10,
		(sc->dev, "Busy Status of functions in group 2: 0x%04x\n",
		 funcstat[26] << 8 | funcstat[27]));
	    LOG(0x10,
		(sc->dev, "Busy Status of functions in group 1: 0x%04x\n",
		 funcstat[28] << 8 | funcstat[29]));
	}

	if (sdshc_use_highspeed_mode && (funcstat[16] & 0x0f) == 0x01) {
	    /* Switch to high speed mode */
	    if (sdshcbus_sd_transaction(sc, 6, 0x80fffff1, 0, 64, 1, funcstat)) {
		MSG((sc->dev, "CMD6 (SWITCH_FUNC, mode 1) failed\n"));
	    } else {
		if ((funcstat[16] & 0x0f) != 0x01) {
		    MSG((sc->dev, "Switching to high speed mode failure\n"));
		} else {
		    sdshcbus_set_highspeed_mode(sc, 1);
		    sc->highspeed_enabled = 1;
		    MSG((sc->dev, "Switched to High Speed mode\n"));
		}

		/* Deselect card */
		sdshcbus_sd_transaction(sc, 7, 0, 0, 0, 0, NULL); /* no resp */
		/* Reget CSD by CMD9 */
		if ((ret = sdshcbus_sd_get_CSD(sc)) != 0) {
		    return ret;
		}
		/* Change new clock */
		sdshcbus_clock_frequency_change(sc, sc->card_max_clock);
		/* Reselect card */
		if (sdshcbus_sd_transaction(sc, 7, sc->rca, 0, 0, 0, NULL)) {
		    MSG((sc->dev, "CMD7 (SELECT_CARD: 0x%08x) failed\n",
			 sc->rca));
		    return ENXIO;
		}
	    }
	}
    }

    sc->card_speed_class = -1;
    /* get SD Status */
    if (sdshcbus_acmd_transaction(sc, 13, 0, 0, 64, 1, sd_status)) {
	MSG((sc->dev, "ACMD13 (SD_STATUS) failed\n"));
    } else {
	LOG(0x10, (sc->dev, "SD_STATUS: 0x"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x;"
		   "%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
		   sd_status[ 0], sd_status[ 1], sd_status[ 2], sd_status[ 3],
		   sd_status[ 4], sd_status[ 5], sd_status[ 6], sd_status[ 7],
		   sd_status[ 8], sd_status[ 9], sd_status[10], sd_status[11],
		   sd_status[12], sd_status[13], sd_status[14], sd_status[15],
		   sd_status[16], sd_status[17], sd_status[18], sd_status[19],
		   sd_status[20], sd_status[21], sd_status[22], sd_status[23],
		   sd_status[24], sd_status[25], sd_status[26], sd_status[27],
		   sd_status[28], sd_status[29], sd_status[30], sd_status[31],
		   sd_status[32], sd_status[33], sd_status[34], sd_status[35],
		   sd_status[36], sd_status[37], sd_status[38], sd_status[39],
		   sd_status[40], sd_status[41], sd_status[42], sd_status[43],
		   sd_status[44], sd_status[45], sd_status[46], sd_status[47],
		   sd_status[48], sd_status[49], sd_status[50], sd_status[51],
		   sd_status[52], sd_status[53], sd_status[54], sd_status[55],
		   sd_status[56], sd_status[57], sd_status[58], sd_status[59],
		   sd_status[60], sd_status[61], sd_status[62], sd_status[63]));
	LOG(0x10, (sc->dev, "DAT_BUS_WIDTH: %s\n",
		   ((sd_status[0] >> 6) & 0x03) == 0 ? "1-bit" :
		   ((sd_status[0] >> 6) & 0x03) == 2 ? "4-bit" :
		   "(reserved)"));
	LOG(0x10, (sc->dev, "SECURED_MODE: %s\n",
		   (sd_status[0] & 0x20) ? "In Secured Mode"
					 : "Not in the mode"));
	tmp = (sd_status[2] << 8) | sd_status[3];
	LOG(0x10, (sc->dev, "SD_CARD_TYPE: %s\n",
		   tmp == 0 ? "Regular SD RD/WR Card" :
		   tmp == 1 ? "SD ROM Card" :
		   "(reserved)"));
	tmp = (sd_status[4] << 24) | (sd_status[5] << 16)
	    | (sd_status[6] << 8) | sd_status[7];
	LOG(0x10, (sc->dev, "SIZE_OF_PROTECTED_AREA: %u\n", tmp));
	LOG(0x10, (sc->dev, "SPEED_CLASS: %s\n",
		   sd_status[8] == 0 ? "class 0" :
		   sd_status[8] == 1 ? "class 2" :
		   sd_status[8] == 2 ? "class 4" :
		   sd_status[8] == 3 ? "class 6" :
		   "(reserved)"));
	if (sd_status[8] <= 3) {
	    sc->card_speed_class = sd_status[8] << 1;
	}
	if (sd_status[9] == 0) {
	    LOG(0x10, (sc->dev, "PERFORMANCE_MOVE: Not Defined\n"));
	} else if (sd_status[9] == 0xff) {
	    LOG(0x10, (sc->dev, "PERFORMANCE_MOVE: Infinity\n"));
	} else {
	    LOG(0x10, (sc->dev, "PERFORMANCE_MOVE: %dMB/s\n", sd_status[9]));
	}
	LOG(0x10, (sc->dev, "AU_SIZE: %s\n",
		   (sd_status[10] >> 4) == 0 ? "Not Defined" :
		   (sd_status[10] >> 4) == 1 ? "16kB" :
		   (sd_status[10] >> 4) == 2 ? "32kB" :
		   (sd_status[10] >> 4) == 3 ? "64kB" :
		   (sd_status[10] >> 4) == 4 ? "128kB" :
		   (sd_status[10] >> 4) == 5 ? "256kB" :
		   (sd_status[10] >> 4) == 6 ? "512kB" :
		   (sd_status[10] >> 4) == 7 ? "1MB" :
		   (sd_status[10] >> 4) == 8 ? "2MB" :
		   (sd_status[10] >> 4) == 9 ? "4MB" :
		   "(reserved)"));
	tmp = (sd_status[11] << 8) | sd_status[12];
	if (tmp == 0) {
	    LOG(0x10, (sc->dev, "ERASE_SIZE: Erase Time-out Calculation is not supported\n"));
	} else {
	    LOG(0x10, (sc->dev, "ERASE_SIZE: %u AU\n", tmp));
	}
	if ((sd_status[13] >> 2) == 0) {
	    LOG(0x10, (sc->dev, "ERASE_TIMEOUT: Erase Time-out Calculation is not supported\n"));
	} else {
	    LOG(0x10, (sc->dev, "ERASE_TIMEOUT: %d sec\n", sd_status[13] >> 2));
	}
	LOG(0x10, (sc->dev, "ERASE_OFFSET: %d sec\n", sd_status[13] & 0x03));
    }

    sc->flag.error_recovery_wo_abort = 0;

    return 0;
}


static int
sdshcbus_sd_get_CSD(struct sdshcbus_softc *sc)
{
    u_int32_t reg, reg2;
    u_int64_t cap;
    u_int32_t tmp;

    if (sdshcbus_sd_transaction(sc, 9, sc->rca, 0, 0, 0, NULL)) {
	/* Timeout? */
	MSG((sc->dev, "Timeout occurred for CMD9; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
#if defined(SD_DEBUG)
    if (1) {
	u_int32_t resp0, resp1, resp2, resp3;

	resp0 = SDSHCBUS_READ4(sc, RESPONSE0);
	resp1 = SDSHCBUS_READ4(sc, RESPONSE1);
	resp2 = SDSHCBUS_READ4(sc, RESPONSE2);
	resp3 = SDSHCBUS_READ4(sc, RESPONSE3);
	LOG(0x10, (sc->dev, "CSD: 0x%08x:%08x:%08x:%08x\n",
		   resp3, resp2, resp1, resp0));
    }
#endif
    reg = SDSHCBUS_READ4(sc, RESPONSE3);
    switch ((reg >> 22) & 0x3) {
      case 0:
	/* CSD Ver 1.0 - for Standard Capacity */
	LOG(0x10, (sc->dev, "CSD structure: Standard Capacity\n"));
	reg2 = reg >> 8;
	LOG(0x10, (sc->dev, "  read access time: %d.%d x 1%s\n",
		   time_value[(reg2 >> 3) & 0x0f] / 10,
		   time_value[(reg2 >> 3) & 0x0f] % 10,
		   (reg2 & 0x07) == 0x0 ? "ns" :
		   (reg2 & 0x07) == 0x1 ? "0ns" :
		   (reg2 & 0x07) == 0x2 ? "00ns" :
		   (reg2 & 0x07) == 0x3 ? "us" :
		   (reg2 & 0x07) == 0x4 ? "0us" :
		   (reg2 & 0x07) == 0x5 ? "00us" :
		   (reg2 & 0x07) == 0x6 ? "ms" :
		   (reg2 & 0x07) == 0x7 ? "0ms" :
		   "**"));
	sc->card_read_access_time
	    = time_value[(reg2 >> 3) & 0x0f] * ipow(10, (reg2 & 0x07));
	reg = SDSHCBUS_READ4(sc, RESPONSE2);
	reg2 = reg >> 24;
	LOG(0x10, (sc->dev, "  max trans rate: %d.%d x 1%s\n",
		   time_value[(reg2 >> 3) & 0x0f] / 10,
		   time_value[(reg2 >> 3) & 0x0f] % 10,
		   (reg2 & 0x07) == 0x0 ? "00kbps" :
		   (reg2 & 0x07) == 0x1 ? "Mbps" :
		   (reg2 & 0x07) == 0x2 ? "0Mbps" :
		   (reg2 & 0x07) == 0x3 ? "00Mbps" :
		   "**"));
	sc->card_max_clock
	    = time_value[(reg2 >> 3) & 0x0f] * ipow(10, (reg2 & 0x07)) * 10;
	LOG(0x10, (sc->dev, "  CCC: 0x%03x\n", reg >> 12));
	sc->block_size = 1 << ((reg >> 8) & 0x0f);
	LOG(0x10, (sc->dev, "  max block len: %dB\n",
		   sc->block_size));
	cap = sc->block_size;
	if (sc->max_block_size < sc->block_size) {
	    sc->block_size = sc->max_block_size;
	}
	if (512 < sc->block_size) {
	    sc->block_size = 512;
	}
	LOG(0x10, (sc->dev, "  partial blocks for read allowed: %s\n",
		   (reg >> 7) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  DSR implemented: %s\n",
		   (reg >> 4) & 0x01 ? "yes" : "no"));
	reg2 = SDSHCBUS_READ4(sc, RESPONSE1);
	reg = ((reg & 0x03) << 10) | (reg2 >> 22);
	LOG(0x10, (sc->dev, "  device size: %d\n", reg));
	cap *= (reg + 1);
	reg = (reg2 >> 7) & 0x7;
	LOG(0x10, (sc->dev, "  device size multiplier: %d\n", reg));
	cap *= (1 << (reg + 2));
	sc->card_capacity = (u_int32_t)(cap >> 9);
	LOG(0x10, (sc->dev, "  erase single block enable: %s\n",
		   (reg2 >> 6) & 0x01 ? "yes" : "no"));
	reg = SDSHCBUS_READ4(sc, RESPONSE0);
	LOG(0x10, (sc->dev, "  erase sector size: %d block\n",
		   (((reg2 & 0x3f) << 1) | (reg >> 31)) + 1));
	LOG(0x10, (sc->dev, "  write protect group size: %d sectors\n",
		   ((reg >> 24) & 0x7f) + 1));
	LOG(0x10, (sc->dev, "  write protect group enable: %s\n",
		   (reg >> 23) & 0x01 ? "yes" : "no"));
	tmp = 1 << ((reg >> 18) & 0x7);
	LOG(0x10, (sc->dev, "  write speed factor: /%d\n", tmp));
	sc->card_write_speed_factor = tmp;
	LOG(0x10, (sc->dev, "  partial blocks for write allowed: %s\n",
		   (reg >> 13) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  file format group: %d\n", (reg >> 7) & 0x01));
	LOG(0x10, (sc->dev, "  permanent write protection: %s\n",
		   (reg >> 5) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  temporary write protection: %s\n",
		   (reg >> 4) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  file format: %s\n",
		   ((reg >> 7) & 0x01) ? "Reserved" :
		   ((reg >> 2) & 0x03) == 0 ? "HDD like (w/ partition table)" :
		   ((reg >> 2) & 0x03) == 1 ? "DOS FAT (w/ boot sector only)" :
		   ((reg >> 2) & 0x03) == 2 ? "Universal File Format" :
		   "Others/Unknown"));
	break;
      case 1:
	/* CSD Ver 2.0 - for High Capacity */
	LOG(0x10, (sc->dev, "CSD structure: Ver 2.00/High Capacity\n"));
	reg2 = reg >> 8;
	LOG(0x10, (sc->dev, "  read access time: %d.%d x 1%s\n",
		   time_value[(reg2 >> 3) & 0x0f] / 10,
		   time_value[(reg2 >> 3) & 0x0f] % 10,
		   (reg2 & 0x07) == 0x0 ? "ns" :
		   (reg2 & 0x07) == 0x1 ? "0ns" :
		   (reg2 & 0x07) == 0x2 ? "00ns" :
		   (reg2 & 0x07) == 0x3 ? "us" :
		   (reg2 & 0x07) == 0x4 ? "0us" :
		   (reg2 & 0x07) == 0x5 ? "00us" :
		   (reg2 & 0x07) == 0x6 ? "ms" :
		   (reg2 & 0x07) == 0x7 ? "0ms" :
		   "**"));
	sc->card_read_access_time
	    = time_value[(reg2 >> 3) & 0x0f] * ipow(10, (reg2 & 0x07));
	reg = SDSHCBUS_READ4(sc, RESPONSE2);
	reg2 = reg >> 24;
	LOG(0x10, (sc->dev, "  max trans rate: %d.%d x 1%s\n",
		   time_value[(reg2 >> 3) & 0x0f] / 10,
		   time_value[(reg2 >> 3) & 0x0f] % 10,
		   (reg2 & 0x07) == 0x0 ? "00kbps" :
		   (reg2 & 0x07) == 0x1 ? "Mbps" :
		   (reg2 & 0x07) == 0x2 ? "0Mbps" :
		   (reg2 & 0x07) == 0x3 ? "00Mbps" :
		   "**"));
	sc->card_max_clock
	    = time_value[(reg2 >> 3) & 0x0f] * ipow(10, (reg2 & 0x07)) * 10;
	reg = SDSHCBUS_READ4(sc, RESPONSE2);
	LOG(0x10, (sc->dev, "  CCC: 0x%03x\n", reg >> 12));
	sc->block_size = 1 << ((reg >> 8) & 0x0f);
	LOG(0x10, (sc->dev, "  max block len: %dB\n",
		   sc->block_size));
	if (sc->max_block_size < sc->block_size) {
	    sc->block_size = sc->max_block_size;
	}
	if (512 < sc->block_size) {
	    sc->block_size = 512;
	}
	LOG(0x10, (sc->dev, "  partial blocks for read allowed: %s\n",
		   (reg >> 7) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  DSR implemented: %s\n",
		   (reg >> 4) & 0x01 ? "yes" : "no"));
	reg2 = SDSHCBUS_READ4(sc, RESPONSE1);
	reg = (reg2 >> 8) & 0x003fffff;
	LOG(0x10, (sc->dev, "  device size: %d\n", reg));
	cap = (reg + 1) << 19;	/* (C_SIZE + 1) * 512K */
	sc->card_capacity = (u_int32_t)(cap >> 9);
	LOG(0x10, (sc->dev, "  erase single block enable: %s\n",
		   (reg2 >> 6) & 0x01 ? "yes" : "no"));
	reg = SDSHCBUS_READ4(sc, RESPONSE0);
	LOG(0x10, (sc->dev, "  erase sector size: %d block\n",
		   (((reg2 & 0x3f) << 1) | (reg >> 31)) + 1));
	LOG(0x10, (sc->dev, "  write protect group size: %d sectors\n",
		   ((reg >> 24) & 0x7f) + 1));
	LOG(0x10, (sc->dev, "  write protect group enable: %s\n",
		   (reg >> 23) & 0x01 ? "yes" : "no"));
	tmp = 1 << ((reg >> 18) & 0x7);
	LOG(0x10, (sc->dev, "  write speed factor: /%d\n", tmp));
	sc->card_write_speed_factor = tmp;
	LOG(0x10, (sc->dev, "  partial blocks for write allowed: %s\n",
		   (reg >> 13) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  file format group: %d\n", (reg >> 7) & 0x01));
	LOG(0x10, (sc->dev, "  permanent write protection: %s\n",
		   (reg >> 5) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  temporary write protection: %s\n",
		   (reg >> 4) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  file format: %s\n",
		   ((reg >> 7) & 0x01) ? "Reserved" :
		   ((reg >> 2) & 0x03) == 0 ? "HDD like (w/ partition table)" :
		   ((reg >> 2) & 0x03) == 1 ? "DOS FAT (w/ boot sector only)" :
		   ((reg >> 2) & 0x03) == 2 ? "Universal File Format" :
		   "Others/Unknown"));
	break;
      default:
	LOG(0x10, (sc->dev, "CSD structure: Unknown\n"));
	break;
    }

    return 0;
}


static int sdshcbus_mmc_bustest_w(struct sdshcbus_softc *sc, int len, void *data);
static int sdshcbus_mmc_bustest_r(struct sdshcbus_softc *sc, int len, void *data);

static int
sdshcbus_card_init_mmc(struct sdshcbus_softc *sc)
{
    u_int32_t reg, reg2, reg3;
    int csd_struct, spec_vers;
    int voltage = sc->card_voltage;
    char *ptr;
    u_int64_t cap;
    u_int32_t tmp;

    sc->card_inserted = SDDEV_MMC;
    sc->card_speed_class = -1;

    if (sdshcbus_mmc_transaction(sc, 0, 0, 0, 0, 0, NULL)) {
	MSG((sc->dev, "Timeout occured for CMD0; giving up\n"));
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	sc->card_inserted = SDDEV_INVALID;
	return 1;
    }

    /* build up voltage range code */
    reg = 0x40000000;	/* for MMC HD support */
    if ((sc->support_voltage & 0x1)) {
	/* contoller support 3.3V */
	reg |= 0x000f8000;
    }
    if ((sc->support_voltage & 0x2)) {
	/* contoller support 3.0V */
	reg |= 0x00f80000;
    }
    if ((sc->support_voltage & 0x4)) {
	/* contoller support 1.8V */
	reg |= 0x00000080;
    }

    LOG(0x10, (sc->dev, "arg of CMD1: 0x%08x\n", reg));
    if (sdshcbus_mmc_transaction(sc, 1, reg, 0, 0, 0, NULL)) {
	MSG((sc->dev, "Timeout occured for CMD1; old card?\n"));
	sdshcbus_reset_lines(sc);
	if (sdshcbus_mmc_transaction(sc, 0, 0, 0, 0, 0, NULL)) {
	    MSG((sc->dev, "Timeout occured for CMD0; giving up\n"));
	    sdshcbus_reset_lines(sc);
	    sdshcbus_card_powerdown(sc);
	    sc->card_inserted = SDDEV_INVALID;
	    return 1;
	}

	/* build up voltage range code */
	reg = 0;	/* for MMC SD support only */
	if ((sc->support_voltage & 0x1)) {
	    /* contoller support 3.3V */
	    reg |= 0x000f8000;
	}
	if ((sc->support_voltage & 0x2)) {
	    /* contoller support 3.0V */
	    reg |= 0x00f80000;
	}
	if ((sc->support_voltage & 0x4)) {
	    /* contoller support 1.8V */
	    reg |= 0x00000080;
	}

	LOG(0x10, (sc->dev, "arg of CMD1: 0x%08x\n", reg));
	if (sdshcbus_mmc_transaction(sc, 1, reg, 0, 0, 0, NULL)) {
	    MSG((sc->dev, "Timeout occured for CMD1; giving up\n"));
	    sdshcbus_reset_lines(sc);
	    sdshcbus_card_powerdown(sc);
	    sc->card_inserted = SDDEV_INVALID;
	    return ENXIO;
	}
    }
    reg2 = SDSHCBUS_READ4(sc, RESPONSE0);
    LOG(0x10, (sc->dev, "resp: 0x%08x\n", reg2));

    reg &= 0x60000000;
    /* check card voltage */
    if ((sc->support_voltage & 0x4) && (reg2 & 0x00000080)) {
	/* contoller support 1.8V, card support 1.65 - 1.95V */
	reg |= 0x00000080;
	voltage = 18;
    } else if ((sc->support_voltage & 0x2) && (reg2 & 0x00020000)) {
	/* contoller support 3.0V, card support 2.9 - 3.0V */
	reg |= 0x00020000;
	voltage = 30;
    } else if ((sc->support_voltage & 0x2) && (reg2 & 0x00040000)) {
	/* contoller support 3.0V, card support 3.0 - 3.1V */
	reg |= 0x00040000;
	voltage = 30;
    } else if ((sc->support_voltage & 0x1) && (reg2 & 0x00100000)) {
	/* contoller support 3.3V, card support 3.2 - 3.3V */
	reg |= 0x00100000;
	voltage = 33;
    } else if ((sc->support_voltage & 0x1) && (reg2 & 0x00200000)) {
	/* contoller support 3.3V, card support 3.3 - 3.4V */
	reg |= 0x00200000;
	voltage = 33;
    } else {
	MSG((sc->dev, "Voltage miss-match; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }

    while (!(reg2 & 0x80000000)) {
	LOG(0x10, (sc->dev, "arg of CMD1: 0x%08x\n", reg));
	if (sdshcbus_mmc_transaction(sc, 1, reg, 0, 0, 0, NULL)) {
	    return ENXIO;
	}
	reg2 = SDSHCBUS_READ4(sc, RESPONSE0);
	LOG(0x10, (sc->dev, "resp: 0x%08x\n", reg2));
    }
    if (((reg2 >> 29) & 0x03) == 0x02) {
	sc->card_inserted = SDDEV_MMCHD;
    }

    /* `SD Bus Voltage changed?' */
    if (voltage != sc->card_voltage) {
	MSG((sc->dev, "changing card voltage to %d.%dV\n",
	     voltage / 10, voltage % 10));
	/* `Clr SD Bus Power' */
	sdshcbus_card_powerdown(sc);
	/* `Set SD Bus Voltage Select' */
	sc->card_voltage = voltage;
	sdshcbus_card_powerup(sc);
    }

    /* Identation by CMD2 */
    if (sdshcbus_mmc_transaction(sc, 2, 0, 0, 0, 0, NULL)) {
	/* Timeout? */
	MSG((sc->dev, "Timeout occurred for CMD2; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
#if defined(SD_DEBUG)
    if (1) {
	u_int32_t resp0, resp1, resp2, resp3;

	resp0 = SDSHCBUS_READ4(sc, RESPONSE0);
	resp1 = SDSHCBUS_READ4(sc, RESPONSE1);
	resp2 = SDSHCBUS_READ4(sc, RESPONSE2);
	resp3 = SDSHCBUS_READ4(sc, RESPONSE3);
	LOG(0x10, (sc->dev, "CID: 0x%08x:%08x:%08x:%08x\n",
		   resp3, resp2, resp1, resp0));
    }
#endif
    reg = SDSHCBUS_READ4(sc, RESPONSE3);
    sc->mid = (reg >> 16) & 0xff;
    ptr = sc->oid;
    if (isprint((*ptr = (reg >> 8) & 0xff))) {
	++ptr;
    }
    if (isprint((*ptr = reg & 0xff))) {
	++ptr;
    }
    *ptr = '\0';
    reg = SDSHCBUS_READ4(sc, RESPONSE2);
    sc->pnm[0] = (char)(reg >> 24);
    sc->pnm[1] = (char)(reg >> 16);
    sc->pnm[2] = (char)(reg >>  8);
    sc->pnm[3] = (char)(reg >>  0);
    reg = SDSHCBUS_READ4(sc, RESPONSE1);
    sc->pnm[4] = (char)(reg >> 24);
    sc->pnm[5] = (char)(reg >> 16);
    sc->pnm[6] = '\0';
    sc->prv[0] = (reg >> 12) & 0x0f;
    sc->prv[1] = (reg >>  8) & 0x0f;
    reg2 = SDSHCBUS_READ4(sc, RESPONSE0);
    LOG(0x08, (sc->dev, "Idented: %s rev %d.%d S/N %u (%d/%02d)\n",
	       sc->pnm, sc->prv[0], sc->prv[1], (reg << 24) | (reg2 >> 8),
	       1997 + (reg2 & 0xf), (reg2 >> 4) & 0xf));

    /* set RCA number by CMD3 */
    sc->rca = (sc->slot_no[1] + 2) << 16;
    if (sdshcbus_mmc_transaction(sc, 3, sc->rca, 0, 0, 0, NULL)) {
	/* Timeout? */
	MSG((sc->dev, "Timeout occurred for CMD3; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
    reg = SDSHCBUS_READ4(sc, RESPONSE0);
    LOG(0x10, (sc->dev, "RCA: 0x%04x,"
	       " COM_CRC_ERROR: %d, ILLEGAL_COMMAND: %d, ERROR: %d,"
	       " CURRENT_STATE: %s, READY_FOR_DATA: %d, APP_CMD: %d,"
	       " AKE_SEQ_ERROR: %d\n",
	       sc->rca >> 16,
	       (reg >> 23) & 0x01, (reg >> 22) & 0x01, (reg >> 19) & 0x01,
	       card_state[(reg >> 9) & 0x0f], (reg >> 8) & 0x01,
	       (reg >> 5) & 0x01, (reg >> 3) & 0x01));

    cap = sc->block_size = 0;
    /* get CSD by CMD9 */
    if (sdshcbus_mmc_transaction(sc, 9, sc->rca, 0, 0, 0, NULL)) {
	/* Timeout? */
	MSG((sc->dev, "Timeout occurred for CMD9; this card is unusable\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }
#if defined(SD_DEBUG)
    if (1) {
	u_int32_t resp0, resp1, resp2, resp3;

	resp0 = SDSHCBUS_READ4(sc, RESPONSE0);
	resp1 = SDSHCBUS_READ4(sc, RESPONSE1);
	resp2 = SDSHCBUS_READ4(sc, RESPONSE2);
	resp3 = SDSHCBUS_READ4(sc, RESPONSE3);
	LOG(0x10, (sc->dev, "CSD: 0x%08x:%08x:%08x:%08x\n",
		   resp3, resp2, resp1, resp0));
    }
#endif
    reg = SDSHCBUS_READ4(sc, RESPONSE3);
    csd_struct = (reg >> 22) & 0x3;
    LOG(0x10, (sc->dev, "CSD structure: %s\n",
	       csd_struct == 0 ? "Ver 1.0" : /* v1.0 - 1.2 */
	       csd_struct == 1 ? "Ver 1.1" : /* v1.4 - 2.2 */
	       csd_struct == 2 ? "Ver 1.2" : /* v3.1 - 3.31 - 4.0 - 4.1 */
	       csd_struct == 3 ? "has EXT_CSD" : /* v4.2 */
	       "Reserved"));
    spec_vers = (reg >> 18) & 0xf;
    LOG(0x10, (sc->dev, "SPEC_VERS = 0x%x, Spec Version: %s\n",
	       spec_vers,
	       spec_vers == 0 ? "v1.0 to 1.2" :
	       spec_vers == 1 ? "v1.4" :
	       spec_vers == 2 ? "v2.0 to 2.2" :
	       spec_vers == 3 ? "v3.1 to 3.31" :
	       spec_vers == 4 ? "v4.0 to 4.2" :
	       "Reserved"));
    switch (csd_struct) {
      case 3:
	/* CSD Ver 1.3 - fall trough, and existing EXT_CSD also */
      case 2:
	/* CSD Ver 1.2 - for System Specification Ver 3.1 to 3.3 */
      case 1:
	/* CSD Ver 1.1 - for System Specification Ver 1.4 to 2.2 */
	reg2 = reg >> 8;
	LOG(0x10, (sc->dev, "  read access time: %d.%d x 1%s\n",
		   time_value[(reg2 >> 3) & 0x0f] / 10,
		   time_value[(reg2 >> 3) & 0x0f] % 10,
		   (reg2 & 0x07) == 0x0 ? "ns" :
		   (reg2 & 0x07) == 0x1 ? "0ns" :
		   (reg2 & 0x07) == 0x2 ? "00ns" :
		   (reg2 & 0x07) == 0x3 ? "us" :
		   (reg2 & 0x07) == 0x4 ? "0us" :
		   (reg2 & 0x07) == 0x5 ? "00us" :
		   (reg2 & 0x07) == 0x6 ? "ms" :
		   (reg2 & 0x07) == 0x7 ? "0ms" :
		   "**"));
	sc->card_read_access_time
	    = time_value[(reg2 >> 3) & 0x0f] * ipow(10, (reg2 & 0x07));
	reg = SDSHCBUS_READ4(sc, RESPONSE2);
	reg2 = reg >> 24;
	LOG(0x10, (sc->dev, "  max trans rate: %d.%d x 1%s\n",
		   time_value[(reg2 >> 3) & 0x0f] / 10,
		   time_value[(reg2 >> 3) & 0x0f] % 10,
		   (reg2 & 0x07) == 0x0 ? "00kbps" :
		   (reg2 & 0x07) == 0x1 ? "Mbps" :
		   (reg2 & 0x07) == 0x2 ? "0Mbps" :
		   (reg2 & 0x07) == 0x3 ? "00Mbps" :
		   "**"));
	sc->card_max_clock
	    = time_value[(reg2 >> 3) & 0x0f] * ipow(10, (reg2 & 0x07)) * 10;
	LOG(0x10, (sc->dev, "  CCC: 0x%03x\n", reg >> 12));
	sc->block_size = 1 << ((reg >> 8) & 0x0f);
	LOG(0x10, (sc->dev, "  max block len: %dB\n",
		   sc->block_size));
	cap = sc->block_size;
	if (sc->max_block_size < sc->block_size) {
	    sc->block_size = sc->max_block_size;
	}
	if (512 < sc->block_size) {
	    sc->block_size = 512;
	}
	LOG(0x10, (sc->dev, "  partial blocks for read allowed: %s\n",
		   (reg >> 7) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  DSR implemented: %s\n",
		   (reg >> 4) & 0x01 ? "yes" : "no"));
	reg2 = SDSHCBUS_READ4(sc, RESPONSE1);
	reg = ((reg & 0x03) << 10) | (reg2 >> 22);
	LOG(0x10, (sc->dev, "  device size: %d\n", reg));
	cap *= (reg + 1);
	reg = (reg2 >> 7) & 0x7;
	LOG(0x10, (sc->dev, "  device size multiplier: %d\n", reg));
	cap *= (1 << (reg + 2));
	sc->card_capacity = (u_int32_t)(cap >> 9);
	reg3 = ((reg2 >> 2) & 0x1f) + 1;
	reg = SDSHCBUS_READ4(sc, RESPONSE0);
	reg3 *= ((((reg2 << 3) & 0x18) | (reg >> 29)) + 1);
	LOG(0x10, (sc->dev, "  size of erasable unit: %d\n", reg3));
	LOG(0x10, (sc->dev, "  write protect group size: %d sectors\n",
		   ((reg >> 24) & 0x1f) + 1));
	LOG(0x10, (sc->dev, "  write protect group enable: %s\n",
		   (reg >> 23) & 0x01 ? "yes" : "no"));
	tmp = 1 << ((reg >> 18) & 0x7);
	LOG(0x10, (sc->dev, "  write speed factor: /%d\n", tmp));
	sc->card_write_speed_factor = tmp;
	LOG(0x10, (sc->dev, "  partial blocks for write allowed: %s\n",
		   (reg >> 13) & 0x01 ? "yes" : "no"));
	if (1 < csd_struct) {
	    LOG(0x10, (sc->dev, "  content protection application:"
		       " %ssupported\n", (reg >> 8) & 0x01 ? "" : "not "));
	}
	LOG(0x10, (sc->dev, "  file format group: %d\n", (reg >> 7) & 0x01));
	LOG(0x10, (sc->dev, "  permanent write protection: %s\n",
		   (reg >> 5) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  temporary write protection: %s\n",
		   (reg >> 4) & 0x01 ? "yes" : "no"));
	LOG(0x10, (sc->dev, "  file format: %s\n",
		   ((reg >> 7) & 0x01) ? "Reserved" :
		   ((reg >> 2) & 0x03) == 0 ? "HDD like (w/ partition table)" :
		   ((reg >> 2) & 0x03) == 1 ? "DOS FAT (w/ boot sector only)" :
		   ((reg >> 2) & 0x03) == 2 ? "Universal File Format" :
		   "Others/Unknown"));
	break;
      default:
	LOG(0x08, (sc->dev, "Unknown CSD structure; Sorry, I gave up\n"));
	sc->card_inserted = SDDEV_INVALID;
	sdshcbus_reset_lines(sc);
	sdshcbus_card_powerdown(sc);
	return ENXIO;
    }

    sdshcbus_print_card_type(sc);

    if (sc->card_max_clock != 0) {
	sdshcbus_clock_frequency_change(sc, sc->card_max_clock);
    }

    /* Addressing card */
    if (sdshcbus_mmc_transaction(sc, 7, sc->rca, 0, 0, 0, NULL)) {
	MSG((sc->dev, "CMD7 (SELECT_CARD: 0x%08x) failed\n", sc->rca));
	return ENXIO;
    }

    /* Set block size */
    int blk_size = sc->block_size;
    if (sdshcbus_mmc_transaction(sc, 16, blk_size, 0, 0, 0, NULL)) {
	MSG((sc->dev, "CMD16 (SET_BLOCKLEN: %d) failed\n", blk_size));
	return ENXIO;
    }

    sc->card_max_buswidth = 1;
    sc->card_buswidth = 1;
    if (4 <= spec_vers) {
	/* card has EXT_CSD; maybe HS-MMC */
	u_int8_t test_data[8];

	sdshcbus_set_bus_width(sc, 4);
	test_data[0] = 0x5a;
	test_data[1] = 0x5a;
	test_data[2] = 0x5a;
	test_data[3] = 0x5a;
	test_data[4] = 0x5a;
	test_data[5] = 0x5a;
	test_data[6] = 0x5a;
	test_data[7] = 0x5a;
	if (!sdshcbus_mmc_transaction(sc, 6, 0x03b70100, 0, 0, 0, NULL)
	    && (1 || !sdshcbus_mmc_bustest_w(sc, 8, test_data))
	    && (1 || !sdshcbus_mmc_bustest_r(sc, 8, test_data))
	    && (1 || test_data[0] == 0xa5)) {
	    sc->card_max_buswidth = 4;
	    sc->card_buswidth = 4;
	} else {
	    sdshcbus_set_bus_width(sc, 1);
	}
    }

    sc->flag.error_recovery_wo_abort = 0;

    return 0;
}


static int
sdshcbus_mmc_bustest_w(struct sdshcbus_softc *sc, int len, void *data)
{
    int cnt;
    u_int8_t *ptr = data;

    sdshcbus_led_control(sc, -1);

    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    /* `Set Block Size Reg' */
    LOG(0x10, (sc->dev, "set BLKSIZE: %d\n", len));
    SDSHCBUS_WRITE2(sc, BLKSIZE, len);

    /* `Set Block Count Reg' */
    LOG(0x10, (sc->dev, "set BLKCOUNT: 1\n"));
    SDSHCBUS_WRITE2(sc, BLKCOUNT, 1);

    /* `Set Argument Reg' */
    LOG(0x10, (sc->dev, "set ARGUMENT: 0\n"));
    SDSHCBUS_WRITE4(sc, ARGUMENT, 0);

    /* `Set Transfer Mode Reg' */
    LOG(0x10, (sc->dev, "set XFERMODE: 0x02\n"));
    SDSHCBUS_WRITE2(sc, XFERMODE, 0x02);

    if (!sc->flag.use_polling) {
	sc->addr = ptr;
	sc->rest_blk_count = 1;
    } else {
	sc->addr = NULL;
	sc->rest_blk_count = 0;
    }

    LOG(0x10, (sc->dev, "Intr status = 0x%04x\n",
	       SDSHCBUS_READ2(sc, NORM_INTR_ST)));

    /* `Set Command Reg' */
    LOG(0x10, (sc->dev, "Issue CMD19\n"));
    SDSHCBUS_WRITE2(sc, COMMAND, sdshcbus_build_cmd(mmc_cmd_attr, 19));

    if (sc->flag.use_polling) {
	/* `Wait for Command Complete' */
	LOG(0x10, (sc->dev, "Polling Command Complete; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if ((sc->result = SDSHCBUS_READ2(sc, NORM_INTR_ST))
					& SDSHC_HSR_NORM_INTR_CMD_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		SDSHCBUS_WRITE2(sc,
			      NORM_INTR_ST, SDSHC_HSR_NORM_INTR_CMD_COMPLETE);
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		LOG(0x10, (sc->dev, "Error occurred\n"));
		sdshcbus_led_control(sc, -1);
		sdshcbus_error_recovery(sc);
		return 1;
	    }
	    SD_SLEEP(sc, PPAUSE,
		     "sdshcbus_cmd_complete_seq", sdshc_polling_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
    } else {
	/* `Wait for Command Complete Int' */
	LOG(0x10, (sc->dev, "Waiting for Command Complete Intr; 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if (sc->status & SDSHCBUS_STAT_COMMAND_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete Int occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		break;
	    }
	    SD_SLEEP(sc,
		     PCATCH, "sdshcbus_cmd_with_data", sdshc_timeout_interval);
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		break;
	    }
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Command Complete timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    if (sc->flag.use_polling) {
	/* `Wait for Buffer Write Ready' */
	LOG(0x10, (sc->dev, "Polling Buffer Write Ready; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if ((SDSHCBUS_READ2(sc, NORM_INTR_ST)
		 & SDSHC_HSR_NORM_INTR_BUFWR_READY)) {
		LOG(0x10, (sc->dev, "Buffer Write Ready; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		SDSHCBUS_WRITE2(sc, NORM_INTR_ST, SDSHC_HSR_NORM_INTR_BUFWR_READY);
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		break;
	    }
	    SD_SLEEP(sc, PCATCH, "sdshcbus_write", sdshc_timeout_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
	if (SDSHC_TIMEOUT_COUNT <= cnt) {
	    MSG((sc->dev, "Buffer Write Ready timeout (#1 block left)\n"));
	    sdshcbus_led_control(sc, -1);
	    return -1;
	}

	/* `Set Block Data' */
	for (cnt = len; 0 < cnt; cnt -= 4, ptr += 4) {
	    LOG(0x10, (sc->dev, "write data: 0x%08x at %p\n",
		       *((u_int32_t *)ptr), ptr));
	    SDSHCBUS_WRITE4(sc, BUFDATA_PORT, *((u_int32_t *)ptr));
	}
	if (cnt < 0) {
	    ptr += cnt;
	}
    }

    /* wait for card is ready */
    LOG(0x10, (sc->dev, "Waiting for card becomes ready...\n"));
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ4(sc, PRESENT_STATE) & SDSHC_HSR_PRESENT_STATE_DAT0)) {
	    LOG(0x08, (sc->dev, "Card is Ready\n"));
	    break;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_write_complete_seq", 1);
	if (0 < cnt) {
	    LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Card ready timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    LOG(0x10, (sc->dev, "transaction done\n"));
    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    sdshcbus_led_control(sc, -1);
    return 0;
}


static int
sdshcbus_mmc_bustest_r(struct sdshcbus_softc *sc, int len, void *data)
{
    int cnt;
    u_int8_t *ptr = data;

    sdshcbus_led_control(sc, -1);

    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    /* `Set Block Size Reg' */
    LOG(0x10, (sc->dev, "set BLKSIZE: 0x%04x\n", len));
    SDSHCBUS_WRITE2(sc, BLKSIZE, len);

    /* `Set Block Count Reg' */
    LOG(0x10, (sc->dev, "set BLKCOUNT: 1\n"));
    SDSHCBUS_WRITE2(sc, BLKCOUNT, 1);

    /* `Set Argument Reg' */
    LOG(0x10, (sc->dev, "set ARGUMENT: 0\n"));
    SDSHCBUS_WRITE4(sc, ARGUMENT, 0);

    /* `Set Transfer Mode Reg' */
    LOG(0x10, (sc->dev, "set XFERMODE: 0x12\n"));
    SDSHCBUS_WRITE2(sc, XFERMODE, 0x12);

    if (!sc->flag.use_polling) {
	sc->addr = ptr;
	sc->rest_blk_count = 1;
    } else {
	sc->addr = NULL;
	sc->rest_blk_count = 0;
    }

    LOG(0x10, (sc->dev, "Intr status = 0x%04x\n",
	       SDSHCBUS_READ2(sc, NORM_INTR_ST)));

    /* `Set Command Reg' */
    LOG(0x10, (sc->dev, "Issue CMD14\n"));
    SDSHCBUS_WRITE2(sc, COMMAND, sdshcbus_build_cmd(mmc_cmd_attr, 14));

    if (sc->flag.use_polling) {
	/* `Wait for Command Complete' */
	LOG(0x10, (sc->dev, "Polling Command Complete; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if ((sc->result = SDSHCBUS_READ2(sc, NORM_INTR_ST))
					& SDSHC_HSR_NORM_INTR_CMD_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		SDSHCBUS_WRITE2(sc,
			      NORM_INTR_ST, SDSHC_HSR_NORM_INTR_CMD_COMPLETE);
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		LOG(0x10, (sc->dev, "Error occurred\n"));
		sdshcbus_led_control(sc, -1);
		sdshcbus_error_recovery(sc);
		return 1;
	    }
	    SD_SLEEP(sc, PPAUSE,
		     "sdshcbus_cmd_complete_seq", sdshc_polling_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
    } else {
	/* `Wait for Command Complete Int' */
	LOG(0x10, (sc->dev, "Waiting for Command Complete Intr; 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if (sc->status & SDSHCBUS_STAT_COMMAND_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete Int occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		break;
	    }
	    SD_SLEEP(sc,
		     PCATCH, "sdshcbus_cmd_with_data", sdshc_timeout_interval);
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		break;
	    }
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Command Complete timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    if (sc->flag.use_polling) {
	/* Wait for Buffer Read Ready */
	LOG(0x10, (sc->dev, "Polling Buffer Read Ready; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if ((SDSHCBUS_READ2(sc, NORM_INTR_ST)
		 & SDSHC_HSR_NORM_INTR_BUFRD_READY)) {
		LOG(0x10, (sc->dev, "Buffer Read Ready; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
			      SDSHC_HSR_NORM_INTR_BUFRD_READY);
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		break;
	    }
	    SD_SLEEP(sc, PCATCH, "sdshcbus_read", sdshc_polling_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
	if (SDSHC_TIMEOUT_COUNT <= cnt) {
	    MSG((sc->dev, "Buffer Read Ready timeout (#1 block left)\n"));
	    sdshcbus_led_control(sc, -1);
	    return -1;
	}

	/* `Get Block Data' */
	for (cnt = len; 4 <= cnt; cnt -= 4, ptr += 4) {
	    *((u_int32_t *)ptr) = SDSHCBUS_READ4(sc, BUFDATA_PORT);
	    LOG(0x10, (sc->dev, "read data: 0x%08x to %p\n",
		       *((u_int32_t *)ptr), ptr));
	}
	if (0 < cnt) {
	    u_int32_t reg;

	    reg = SDSHCBUS_READ4(sc, BUFDATA_PORT);

	    if (2 <= cnt) {
		*((u_int16_t *)ptr) = (u_int16_t)reg;
		LOG(0x10, (sc->dev, "read data: 0x%04x to %p\n",
			   (reg & 0xffff), ptr));
		ptr += 2;
		cnt -= 2;
		reg >>= 16;
	    }
	    if (1 <= cnt) {
		*ptr++ = (u_int8_t)reg;
		LOG(0x10, (sc->dev, "read data: 0x%02x to %p\n",
			   (reg & 0xff), ptr));
	    }
	}
    }

    /* wait for card is ready */
    LOG(0x10, (sc->dev, "Waiting for card becomes ready...\n"));
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ4(sc, PRESENT_STATE) & SDSHC_HSR_PRESENT_STATE_DAT0)) {
	    LOG(0x08, (sc->dev, "Card is Ready\n"));
	    break;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_write_complete_seq", 1);
	if (0 < cnt) {
	    LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Card ready timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    LOG(0x10, (sc->dev, "transaction done\n"));
    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    sdshcbus_led_control(sc, -1);

    return 0;
}


static void
sdshcbus_set_bus_width(struct sdshcbus_softc *sc, int width)
{
    u_int8_t reg;

    reg = SDSHCBUS_READ1(sc, HOST_CTL);
    reg &= ~0x02;
    reg |= ((width == 4) << 1);
    SDSHCBUS_WRITE1(sc, HOST_CTL, reg);
}


static void
sdshcbus_change_bus_width(struct sdshcbus_softc *sc, int width)
{
    int prev_width = sc->card_buswidth;

    if (width == prev_width) {
	LOG(0x08, (sc->dev, "Bus width is same; nothing is needed\n"));
	return;
    }

    /*
    sdshcbus_disable_intr(sc);
    */

    if (is_card_sdio(sc->card_inserted)) {
	/*
	 * Mask Card interrupt in Card
	 */
	/* XXX Set `IENM' of CCCR to 0 by CMD52 */
    }

    /*
     * Change bit mode for Card
     */
    if (sdshcbus_acmd_transaction(sc, 6, width >> 1, 0, 0, 0, NULL)) {
	MSG((sc->dev, "ACMD6 (SET_BUS_WIDTH) failed\n"));
	return;
    }
    if (is_card_sdio(sc->card_inserted)) {
	/* XXX Set `Bus Width' of Bus Interface Control reg in CCCR by CMD52 */
	if (sdshcbus_sd_transaction(sc, 52, 0, 0, 0, 0, NULL)) {
	    MSG((sc->dev, "ACMD6 (SET_BUS_WIDTH) failed\n"));
	    if (sdshcbus_acmd_transaction(sc, 6, prev_width >> 1, 0, 0, 0, NULL)) {
		MSG((sc->dev, "ACMD6 (SET_BUS_WIDTH) failed; giving up\n"));
		return;
	    }
	}
    }

    /*
     * Change bit mode for Host
     */
    sdshcbus_set_bus_width(sc, width);

    if (is_card_sdio(sc->card_inserted)) {
	/*
	 * Enable Card interrupt in Card
	 */
	/* XXX Set `IENM' of CCCR to 0 by CMD52 */
    }

    /*
    sdshcbus_enable_intr(sc);
    */

    sc->card_buswidth = width;
    LOG(0x08, (sc->dev, "card bus width changed to %d-bit\n", width));
}


static void
sdshcbus_set_highspeed_mode(struct sdshcbus_softc *sc, int enable)
{
    u_int8_t reg;

    reg = SDSHCBUS_READ1(sc, HOST_CTL);
    reg &= ~0x04;
    reg |= ((enable != 0) << 2);
    SDSHCBUS_WRITE1(sc, HOST_CTL, reg);
}


/* timeout: [ms] */
static void
sdshcbus_set_data_timeout(struct sdshcbus_softc *sc, u_int32_t timeout)
{
    int cnt, divisor;

    /* `Calculate a divisor for detecting timeout' */
    timeout *= sc->timeout_clock_freq;
    timeout >>= 13;
    for (cnt = 0, divisor = 1; cnt < 14; cnt++, divisor <<= 1) {
	if (timeout < divisor) {
	    break;
	}
    }
    if (0x0e < divisor) {
	divisor = 0x0e;
    }

    /* `Set timeout detection timer' */
    LOG(0x10, (sc->dev, "Data timeout = 0x%x (%dns)\n",
	       divisor, (1 << (1 + 13)) * sc->timeout_clock_freq));
    SDSHCBUS_WRITE1(sc, TIMEOUT_CTL, divisor);
}


static u_int32_t
sdshcbus_build_cmd(struct cmd_type *cmd_attr, int index)
{
    u_int32_t cmd;

    cmd = (index << 8) | ((index == 12 || index == 52 ? 0x3 : 0x0) << 6);
    cmd |= ((cmd_attr[index].need_transfer << 5) | cmd_attr[index].resp_type);
    return cmd;
}


static int
sdshcbus_transaction_wo_data(struct sdshcbus_softc *sc,
			   u_int32_t cmd, u_int32_t arg,
			   int with_trans_complete)
{
    int cnt;

    sdshcbus_led_control(sc, -1);

    /* `Check Command Inhibit (CMD)' */
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	if ((SDSHCBUS_READ4(sc, PRESENT_STATE)
	     & SDSHC_HSR_PRESENT_STATE_CMD_INH_CMD) == 0) {
	    break;
	}
	SD_SLEEP(sc, PPAUSE, "sdshcbus_trans_wo_data", sdshc_polling_interval);
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Command Inhibit (CMD) timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    /* `Issue the command with the busy?' */
    if ((cmd & 0x03) == 0x03) {
	/* `Issue Abort Command?' */
	if ((cmd & 0x00c0) != 0x00c0) {
	    /* `Check Command Inhibit (DAT)' */
	    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		if ((SDSHCBUS_READ4(sc, PRESENT_STATE)
		     & SDSHC_HSR_PRESENT_STATE_CMD_INH_DAT) == 0) {
		    break;
		}
	    }
	    if (SDSHC_TIMEOUT_COUNT <= cnt) {
		MSG((sc->dev, "Command Inhibit (DAT) timeout\n"));
		sdshcbus_led_control(sc, -1);
		return -1;
	    }
	}
    }

    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    /* `Set Argument Reg' */
    SDSHCBUS_WRITE4(sc, ARGUMENT, arg);

    /* `Set Command Reg' */
    if (!sc->flag.in_error_recovery) {
	sc->result = 0;
	sc->error = 0;
    }
    SDSHCBUS_WRITE2(sc, COMMAND, cmd);
    LOG(0x10, (sc->dev, "cmd 0x%04x sent\n", cmd));

    if (sc->flag.in_error_recovery) {
	LOG(0x08, (sc->dev, "recovery in progress;"
		   " command complete sequence canceled\n"));
	sdshcbus_led_control(sc, -1);
	return 0;
    }
    return sdshcbus_command_complete_sequence(sc, with_trans_complete);
}


/*
 * `Command Complete Sequence'
 */
static int
sdshcbus_command_complete_sequence(struct sdshcbus_softc *sc,
				 int with_trans_complete)
{
    int err = 0, cnt;

    if (sc->flag.use_polling) {
	/* `Wait for Command Complete' */
	LOG(0x10, (sc->dev, "Polling Command Complete; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	sdshcbus_enable_intr_polling(sc);
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if ((sc->result = SDSHCBUS_READ2(sc, NORM_INTR_ST))
					& SDSHC_HSR_NORM_INTR_CMD_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
					SDSHC_HSR_NORM_INTR_CMD_COMPLETE);
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		LOG(0x10, (sc->dev, "Error occurred\n"));
		sdshcbus_error_recovery(sc);
		err = 1;
		break;
	    }
	    SD_SLEEP(sc,
		    PPAUSE, "sdshcbus_cmd_complete_seq", sdshc_polling_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
	sdshcbus_disable_intr(sc);
    } else {
	/* `Wait for Command Complete Int' */
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if (sc->status & SDSHCBUS_STAT_COMMAND_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete; status = 0x%04x\n", sc->result));
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		err = 1;
		break;
	    }
	    SD_SLEEP(sc,
		    PCATCH, "sdshcbus_cmd_complete_seq", sdshc_timeout_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Command Complete timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    /* `Get Response Data' - do nothing here */

    /* `Command with Transfer Complete int?' */
    if (!err && with_trans_complete) {
	if (sc->flag.use_polling) {
	    /* `Wait for Transfer Complete' */
	    LOG(0x10, (sc->dev, "Polling Transfer Complete; intr = 0x%04x\n",
		       SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	    sdshcbus_enable_intr_polling(sc);
	    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		if ((sc->result = SDSHCBUS_READ2(sc, NORM_INTR_ST)
					& SDSHC_HSR_NORM_INTR_TRANS_COMPLETE)) {
		    LOG(0x10, (sc->dev,
			       "Transfer Complete occurred; intr = 0x%04x\n",
			       SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
					SDSHC_HSR_NORM_INTR_TRANS_COMPLETE);
		    break;
		}
		SD_SLEEP(sc, PPAUSE,
			 "sdshcbus_cmd_complete_seq", sdshc_polling_interval);
		sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
		if (sc->error != 0) {
		    LOG(0x10, (sc->dev, "Error occurred\n"));
		    sdshcbus_error_recovery(sc);
		    err = 1;
		    break;
		}
		if (0 < cnt) {
		    LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
		}
	    }
	    sdshcbus_disable_intr(sc);
	} else {
	    /* `Wait for Transfer Complete Int' */
	    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		if (sc->status & SDSHCBUS_STAT_TRANSFER_COMPLETE) {
		    LOG(0x10, (sc->dev,
			       "Transfer Complete; status = 0x%04x\n",
			       sc->result));
		    break;
		}
		SD_SLEEP(sc, PCATCH,
			 "sdshcbus_cmd_complete_seq", sdshc_timeout_interval);
		sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
		if (sc->error != 0) {
		    err = 1;
		    break;
		}
		if (0 < cnt) {
		    LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
		}
	    }
	}
	if (SDSHC_TIMEOUT_COUNT <= cnt) {
	    MSG((sc->dev, "Transfer Complete timeout\n"));
	    sdshcbus_led_control(sc, -1);
	    return -1;
	}
    }

    /* `Check Response Data' - do nothing here */

    sdshcbus_led_control(sc, -1);

    return err;
}


static int
sdshcbus_transaction_with_data(struct sdshcbus_softc *sc,
			  u_int32_t cmd, u_int32_t arg,
			  int write, int blk_size, int num_blk, void *data)
{
    int cnt;
    u_int32_t reg;
    int multi_trans = 0, infinite_trans = 0, auto_cmd12 = 0;
    u_int8_t *ptr = data;

    /* check data size parameters */
    if (blk_size == 0 || num_blk == 0) {
	/* data size is 0; transfer is not needed */
	return 0;
    }

    sdshcbus_led_control(sc, -1);

    multi_trans = (1 < num_blk);
    if (sc->max_block_size < blk_size) {
	MSG((sc->dev,
	     "Block size (%d) is wrong; Max Block Length is %dB\n",
	     blk_size, sc->max_block_size));
    }
    if (SDSHC_MAX_BLOCK_COUNT < num_blk) {
	infinite_trans = 1;
    } else {
	if (multi_trans) {
	    auto_cmd12 = 1;
	}
    }

    LOG(0x10, (sc->dev,
	       "write: %d, multi_trans: %d, infinite: %d, Auto CMD12: %d\n",
	       write, multi_trans, infinite_trans, auto_cmd12));

    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    /* `Set Block Size Reg' */
    LOG(0x10, (sc->dev, "set BLKSIZE: 0x%04x\n", blk_size));
    SDSHCBUS_WRITE2(sc, BLKSIZE, blk_size);

    /* `Set Block Count Reg' */
    LOG(0x10, (sc->dev, "set BLKCOUNT: 0x%04x\n", num_blk));
    SDSHCBUS_WRITE2(sc, BLKCOUNT, num_blk);

    /* `Set Argument Reg' */
    LOG(0x10, (sc->dev, "set ARGUMENT: 0x%08x\n", arg));
    SDSHCBUS_WRITE4(sc, ARGUMENT, arg);

    /* `Set Transfer Mode Reg' */
    reg = ((multi_trans << 5)
	   | ((write == 0) << 4)
	   | (auto_cmd12 << 2)
	   | ((!infinite_trans) << 1));
    LOG(0x10, (sc->dev, "set XFERMODE: 0x%04x\n", reg));
    SDSHCBUS_WRITE2(sc, XFERMODE, reg);

    if (!sc->flag.use_polling) {
	sc->addr = ptr;
	sc->rest_blk_count = num_blk;
    } else {
	sc->addr = NULL;
	sc->rest_blk_count = 0;
    }

    LOG(0x10, (sc->dev, "Intr status = 0x%04x\n",
	       SDSHCBUS_READ2(sc, NORM_INTR_ST)));

    /* `Set Command Reg' */
    SDSHCBUS_WRITE2(sc, COMMAND, cmd);

    if (sc->flag.use_polling) {
	/* `Wait for Command Complete' */
	LOG(0x10, (sc->dev, "Polling Command Complete; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if ((sc->result = SDSHCBUS_READ2(sc, NORM_INTR_ST))
					& SDSHC_HSR_NORM_INTR_CMD_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		SDSHCBUS_WRITE2(sc,
			      NORM_INTR_ST, SDSHC_HSR_NORM_INTR_CMD_COMPLETE);
		break;
	    }
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		LOG(0x10, (sc->dev, "Error occurred\n"));
		sdshcbus_led_control(sc, -1);
		sdshcbus_error_recovery(sc);
		return 1;
	    }
	    SD_SLEEP(sc, PPAUSE,
		     "sdshcbus_cmd_complete_seq", sdshc_polling_interval);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
    } else {
	/* `Wait for Command Complete Int' */
	LOG(0x10, (sc->dev, "Waiting for Command Complete Intr; 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if (sc->status & SDSHCBUS_STAT_COMMAND_COMPLETE) {
		LOG(0x10, (sc->dev,
			   "Command Complete Int occurred; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		break;
	    }
	    SD_SLEEP(sc,
		     PCATCH, "sdshcbus_cmd_with_data", sdshc_timeout_interval);
	    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
	    if (sc->error != 0) {
		break;
	    }
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Command Complete timeout\n"));
	sdshcbus_led_control(sc, -1);
	return -1;
    }

    /* `Get Response Data' - do nothing here */

    if (sc->flag.use_polling) {
	/* `Write or Read?' */
	if (write) {
	    do {
		/* `Wait for Buffer Write Ready' */
		LOG(0x10, (sc->dev,
			   "Polling Buffer Write Ready; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		    if ((SDSHCBUS_READ2(sc, NORM_INTR_ST)
			 & SDSHC_HSR_NORM_INTR_BUFWR_READY)) {
			LOG(0x10, (sc->dev,
				   "Buffer Write Ready; intr = 0x%04x\n",
				   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
			SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
				   SDSHC_HSR_NORM_INTR_BUFWR_READY);
			break;
		    }
		    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
		    if (sc->error != 0) {
			break;
		    }
		    SD_SLEEP(sc,
			     PCATCH, "sdshcbus_write", sdshc_timeout_interval);
		    if (0 < cnt) {
			LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
		    }
		}
		if (SDSHC_TIMEOUT_COUNT <= cnt) {
		    MSG((sc->dev,
			 "Buffer Write Ready timeout (#%d block(s) left)\n",
			 num_blk));
		    sdshcbus_led_control(sc, -1);
		    return -1;
		}

		/* `Set Block Data' */
		for (cnt = blk_size; 0 < cnt; cnt -= 4, ptr += 4) {
		    LOG(0x10, (sc->dev, "write data: 0x%08x at %p\n",
			       *((u_int32_t *)ptr), ptr));
		    SDSHCBUS_WRITE4(sc, BUFDATA_PORT, *((u_int32_t *)ptr));
		}
		if (cnt < 0) {
		    ptr += cnt;
		}

		/* `More Blocks?' */
	    } while (0 < --num_blk);
	} else {
	    /* read */

	    do {
		/* Wait for Buffer Read Ready */
		LOG(0x10, (sc->dev,
			   "Polling Buffer Read Ready; intr = 0x%04x\n",
			   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		    if ((SDSHCBUS_READ2(sc, NORM_INTR_ST)
			 & SDSHC_HSR_NORM_INTR_BUFRD_READY)) {
			LOG(0x10, (sc->dev,
				   "Buffer Read Ready; intr = 0x%04x\n",
				   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
			SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
				   SDSHC_HSR_NORM_INTR_BUFRD_READY);
			break;
		    }
		    sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
		    if (sc->error != 0) {
			break;
		    }
		    SD_SLEEP(sc,
			     PCATCH, "sdshcbus_read", sdshc_polling_interval);
		    if (0 < cnt) {
			LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
		    }
		}
		if (SDSHC_TIMEOUT_COUNT <= cnt) {
		    MSG((sc->dev,
			 "Buffer Read Ready timeout (#%d block(s) left)\n",
			 num_blk));
		    sdshcbus_led_control(sc, -1);
		    return -1;
		}

		/* `Get Block Data' */
		for (cnt = blk_size; 4 <= cnt; cnt -= 4, ptr += 4) {
		    *((u_int32_t *)ptr) = SDSHCBUS_READ4(sc, BUFDATA_PORT);
		    LOG(0x10, (sc->dev, "read data: 0x%08x to %p\n",
			       *((u_int32_t *)ptr), ptr));
		}
		if (0 < cnt) {
		    u_int32_t reg;

		    reg = SDSHCBUS_READ4(sc, BUFDATA_PORT);

		    if (2 <= cnt) {
			*((u_int16_t *)ptr) = (u_int16_t)reg;
			LOG(0x10, (sc->dev, "read data: 0x%04x to %p\n",
				   (reg & 0xffff), ptr));
			ptr += 2;
			cnt -= 2;
			reg >>= 16;
		    }
		    if (1 <= cnt) {
			*ptr++ = (u_int8_t)reg;
			LOG(0x10, (sc->dev, "read data: 0x%02x to %p\n",
				   (reg & 0xff), ptr));
		    }
		}

		/* `More Blocks?' */
	    } while (0 < --num_blk);
	}
    }

    /* Single Block Transfer? Multi Block Transfer? Infinite Block Transfer? */
    if (!infinite_trans) {
	/* Single or Multi Block Transfer */

	if (sc->flag.use_polling) {
	    /* `Wait for Transfer Complete' */
	    LOG(0x10, (sc->dev, "Polling Transfer Complete; intr = 0x%04x\n",
		   SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		if ((sc->result = SDSHCBUS_READ2(sc, NORM_INTR_ST))
					& SDSHC_HSR_NORM_INTR_TRANS_COMPLETE) {
		    LOG(0x10, (sc->dev,
			       "Transfer Complete occurred; intr = 0x%04x\n",
			       SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
					SDSHC_HSR_NORM_INTR_TRANS_COMPLETE);
		    break;
		}
		sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
		if (sc->error != 0) {
		    LOG(0x10, (sc->dev, "Error occurred\n"));
		    sdshcbus_error_recovery(sc);
		    sdshcbus_led_control(sc, -1);
		    return 1;
		}
		SD_SLEEP(sc, PPAUSE,
			 "sdshcbus_cmd_complete_seq", sdshc_polling_interval);
		if (0 < cnt) {
		    LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
		}
	    }
	} else {
	    /* `Wait for Transfer Complete Int' */
	    LOG(0x10, (sc->dev, "Waiting for Transfer Complete Intr; 0x%04x\n",
		       SDSHCBUS_READ2(sc, NORM_INTR_ST)));
	    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
		if (sc->status & SDSHCBUS_STAT_TRANSFER_COMPLETE) {
		    LOG(0x10, (sc->dev,
			     "Transfer Complete Int occurred; intr = 0x%04x\n",
			       SDSHCBUS_READ2(sc, NORM_INTR_ST)));
		    break;
		}
		sc->error = SDSHCBUS_READ2(sc, ERR_INTR_ST);
		if (sc->error != 0) {
		    break;
		}
		SD_SLEEP(sc, PCATCH,
			 "sdshcbus_cmd_with_data", sdshc_timeout_interval);
	    }
	}
	if (SDSHC_TIMEOUT_COUNT <= cnt) {
	    MSG((sc->dev, "Transfer Complete timeout\n"));
	    sdshcbus_led_control(sc, -1);
	    return -1;
	}
    } else {
	/* Infinite Block Transfer */
	if (write ? sdshcbus_sync_abort(sc)
		  : sdshcbus_async_abort(sc)) {
	    sdshcbus_led_control(sc, -1);
	    return -1;
	}
    }

    if (write) {
	/* wait for exiting programming state */
	LOG(0x10, (sc->dev, "Waiting for finishing programming state...\n"));
	for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	    if (sdshcbus_sd_transaction(sc, 13, sc->rca, 0, 0, 0, NULL)) {
		MSG((sc->dev,
		     "Timeout occured on CMD13; cannot get card state\n"));
		sdshcbus_led_control(sc, -1);
		return -1;
	    }
	    reg = SDSHCBUS_READ4(sc, RESPONSE0);
	    if (reg & 0x100) {
		LOG(0x08, (sc->dev, "Ready for next data\n"));
		break;
	    }
	    SD_SLEEP(sc, PPAUSE, "sdshcbus_write_complete_seq", 1);
	    if (0 < cnt) {
		LOG(0x10, (sc->dev, "wait cnt = %d\n", cnt));
	    }
	}
	if (SDSHC_TIMEOUT_COUNT <= cnt) {
	    MSG((sc->dev, "Write Complete timeout\n"));
	    sdshcbus_led_control(sc, -1);
	    return -1;
	}
    }

    /* `Check Response Data' - do nothing here */

    LOG(0x10, (sc->dev, "transaction done\n"));
    SDSHCBUS_WRITE2(sc, NORM_INTR_ST,
		  (SDSHC_HSR_NORM_INTR_CMD_COMPLETE
		   | SDSHC_HSR_NORM_INTR_TRANS_COMPLETE));
    sc->status &= ~(SDSHCBUS_STAT_COMMAND_COMPLETE
		    | SDSHCBUS_STAT_TRANSFER_COMPLETE);

    sdshcbus_led_control(sc, -1);
    return 0;
}


static int
sdshcbus_transaction(struct sdshcbus_softc *sc,
		   struct cmd_type *cmd_attr,
		   u_int32_t cmd_index, u_int32_t arg,
		   int write, int blk_size, int num_blk, void *data)
{
    u_int32_t cmd;
    int with_trans_complete;

    if (cmd_attr[cmd_index].resp_type == CMD_RESERVED) {
	return -1;
    }
    cmd = sdshcbus_build_cmd(cmd_attr, cmd_index);
    if (cmd_index == 0) {
	LOG(0x10, (sc->dev, "Set Bus Width to 1-bit (default)\n"));
	sdshcbus_set_bus_width(sc, 1);
    }
    LOG(0x10, (sc->dev, "Issue %sCMD%d\n", cmd_attr == sd_acmd_attr ? "A" : "",
	       cmd_index));
    if (cmd_attr[cmd_index].need_transfer) {
	return (sdshcbus_transaction_with_data(sc, cmd, arg, write,
					     blk_size, num_blk, data) != 0);
    } else {
	with_trans_complete = (cmd_attr[cmd_index].resp_type == CMD_R1b
			       || cmd_attr[cmd_index].resp_type == CMD_R5b);
	return (sdshcbus_transaction_wo_data(sc, cmd, arg,
					   with_trans_complete) != 0);
    }
}


static int
sdshcbus_acmd_transaction(struct sdshcbus_softc *sc,
			  u_int32_t cmd_index, u_int32_t arg,
			  int write, int blk_size, int num_blk, void *data)
{
    if (sd_acmd_attr[cmd_index].resp_type == CMD_RESERVED) {
	return -1;
    }

    LOG(0x10, (sc->dev, "Issue CMD55\n"));
    if (sdshcbus_transaction_wo_data(sc, sdshcbus_build_cmd(sd_cmd_attr, 55),
				   sc->rca, 0)) {
	return 1;
    }

    return sdshcbus_transaction(sc, sd_acmd_attr, cmd_index, arg,
			      write, blk_size, num_blk, data);
}


/*
 * Abort Transaction
 */
static int
sdshcbus_async_abort(struct sdshcbus_softc *sc)
{
    u_int16_t cmd;

    /* `Issue Abort Command' */
    if (sc->card_inserted == SDDEV_SDIO) {
	cmd = sdshcbus_build_cmd(sd_cmd_attr, 52);	/* CMD52 */
    } else {
	cmd = sdshcbus_build_cmd(sd_cmd_attr, 12);	/* CMD12 */
    }
    if (sdshcbus_transaction_wo_data(sc, cmd, 0, 0)) {
	return 1;
    }
    if (sc->flag.in_error_recovery) {
	return 0;
    }

    /* `Set Software Reset For DAT line (DR) and CMD line (CR)' */
    return sdshcbus_reset_lines(sc);
}


static int
sdshcbus_sync_abort(struct sdshcbus_softc *sc)
{
    int cnt;
    u_int8_t reg;

    /* `Set Stop at Block Gap Request' */
    reg = SDSHCBUS_READ1(sc, BLKGAP_CTL);
    reg |= 0x01;
    SDSHCBUS_WRITE1(sc, BLKGAP_CTL, reg);

    /* `Wait for Transfer Complete Int' */
    for (cnt = 0; cnt < SDSHC_TIMEOUT_COUNT; cnt++) {
	SD_SLEEP(sc, PCATCH, "sdshcbus_sync_abort", sdshc_timeout_interval);
	if ((sc->result & SDSHC_HSR_NORM_INTR_TRANS_COMPLETE)) {
	    break;
	}
	if (sc->error != 0) {
	    break;
	}
    }
    if (SDSHC_TIMEOUT_COUNT <= cnt) {
	MSG((sc->dev, "Transfer Complete timeout\n"));
	return -1;
    }

    return sdshcbus_async_abort(sc);
}

/****/

static int
sdshcbus_open(struct cdev *dev, int flag, int fmt, struct thread *td)
{
    int unit = SDSHCBUS_UNIT(dev);
    struct sdshcbus_softc *sc = SDSHCBUS_SOFTC(unit);

    if (sc == NULL) {
	return ENXIO;
    }

    /* Disallow multiple opens */
    if (sc->status & SDSHCBUS_STAT_OPENED) {
	return EBUSY;
    }

    device_busy(devclass_get_device(sdshcbus_devclass, unit));

    sc->status |= SDSHCBUS_STAT_OPENED;

    return 0;
}


static int
sdshcbus_close(struct cdev *dev, int flag, int fmt, struct thread *td)
{
    int unit = SDSHCBUS_UNIT(dev);
    struct sdshcbus_softc *sc = SDSHCBUS_SOFTC(unit);

    device_unbusy(devclass_get_device(sdshcbus_devclass, unit));
    sc->status &= ~SDSHCBUS_STAT_OPENED;

    return 0;
}


static int
sdshcbus_read(struct cdev *dev, struct uio *uio, int flag)
{
/*
    int unit = SDSHCBUS_UNIT(dev);
    struct sdshcbus_softc *sc = SDSHCBUS_SOFTC(unit);
*/

    return ENODEV;
}


static int
sdshcbus_write(struct cdev *dev, struct uio *uio, int flag)
{
/*
    int unit = SDSHCBUS_UNIT(dev);
    struct sdshcbus_softc *sc = SDSHCBUS_SOFTC(unit);
*/

    return ENODEV;
}


/*
static int
sdshcbus_poll(struct cdev *dev, int events, struct thread *td)
{
}


static void
sdshcbus_purge(struct cdev *dev)
{
}
*/
